"""scrapli.driver.base_driver"""
import os
import re
import sys
from logging import Logger, getLogger
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Tuple, Union

from scrapli.channel import CHANNEL_ARGS
from scrapli.exceptions import UnsupportedPlatform
from scrapli.helper import _find_transport_plugin, resolve_ssh_config, resolve_ssh_known_hosts
from scrapli.transport import (
    SYSTEM_SSH_TRANSPORT_ARGS,
    TELNET_TRANSPORT_ARGS,
    SystemSSHTransport,
    TelnetTransport,
    Transport,
)

TRANSPORT_CLASS: Dict[str, Callable[..., Transport]] = {
    "system": SystemSSHTransport,
    "telnet": TelnetTransport,
}
TRANSPORT_ARGS: Dict[str, Tuple[str, ...]] = {
    "system": SYSTEM_SSH_TRANSPORT_ARGS,
    "telnet": TELNET_TRANSPORT_ARGS,
}
TRANSPORT_BASE_ARGS = (
    "host",
    "port",
    "keepalive",
    "keepalive_interval",
    "keepalive_type",
    "keepalive_pattern",
    "timeout_socket",
    "timeout_transport",
    "timeout_exit",
)
ASYNCIO_TRANSPORTS = ("asyncssh",)


class ScrapeBase:
    def __init__(
        self,
        host: str = "",
        port: int = 22,
        auth_username: str = "",
        auth_password: str = "",
        auth_private_key: str = "",
        auth_private_key_passphrase: str = "",
        auth_strict_key: bool = True,
        auth_bypass: bool = False,
        timeout_socket: int = 5,
        timeout_transport: int = 10,
        timeout_ops: float = 30,
        timeout_exit: bool = True,
        keepalive: bool = False,
        keepalive_interval: int = 30,
        keepalive_type: str = "network",
        keepalive_pattern: str = "\005",
        comms_prompt_pattern: str = r"^[a-z0-9.\-@()/:]{1,48}[#>$]\s*$",
        comms_return_char: str = "\n",
        comms_ansi: bool = False,
        ssh_config_file: Union[str, bool] = False,
        ssh_known_hosts_file: Union[str, bool] = False,
        on_open: Optional[Callable[..., Any]] = None,
        on_close: Optional[Callable[..., Any]] = None,
        transport: str = "system",
        transport_options: Optional[Dict[str, Any]] = None,
    ):
        r"""
        Scrape Object

        Scrape is the base class for NetworkDriver, and subsequent platform specific drivers (i.e.
        IOSXEDriver). Scrape can be used on its own and offers a semi-pexpect like experience in
        that it doesn't know or care about privilege levels, platform types, and things like that.

        *Note* most arguments passed to Scrape do not actually get assigned to the scrape object
        itself, but instead are used to construct the Transport and Channel classes that Scrape
        relies on, see Transport and Channel docs for details.

        Args:
            host: host ip/name to connect to
            port: port to connect to
            auth_username: username for authentication
            auth_private_key: path to private key for authentication
            auth_private_key_passphrase: passphrase for decrypting ssh key if necessary
            auth_password: password for authentication
            auth_strict_key: strict host checking or not -- applicable for system ssh driver only
            auth_bypass: bypass ssh key or password auth for devices without authentication, or that
                have auth prompts after ssh session establishment. Currently only supported on
                system transport; ignored on other transports
            timeout_socket: timeout for establishing socket in seconds
            timeout_transport: timeout for ssh|telnet transport in seconds
            timeout_ops: timeout for ssh channel operations
            timeout_exit: True/False close transport if timeout encountered. If False and keepalives
                are in use, keepalives will prevent program from exiting so you should be sure to
                catch Timeout exceptions and handle them appropriately
            keepalive: whether or not to try to keep session alive
            keepalive_interval: interval to use for session keepalives
            keepalive_type: network|standard -- 'network' sends actual characters over the
                transport channel. This is useful for network-y type devices that may not support
                'standard' keepalive mechanisms. 'standard' attempts to use whatever 'standard'
                keepalive mechanisms are available in the selected transport mechanism. Check the
                transport documentation for details on what is supported and/or how it is
                implemented for any given transport driver
            keepalive_pattern: pattern to send to keep network channel alive. Default is
                u'\005' which is equivalent to 'ctrl+e'. This pattern moves cursor to end of the
                line which should be an innocuous pattern. This will only be entered *if* a lock
                can be acquired. This is only applicable if using keepalives and if the keepalive
                type is 'network'
            comms_prompt_pattern: raw string regex pattern -- preferably use `^` and `$` anchors!
                this is the single most important attribute here! if this does not match a prompt,
                scrapli will not work!
                IMPORTANT: regex search uses multi-line + case insensitive flags. multi-line allows
                for highly reliably matching for prompts however we do NOT strip trailing whitespace
                for each line, so be sure to add '\\s?' or similar if your device needs that. This
                should be mostly sorted for you if using network drivers (i.e. `IOSXEDriver`).
                Lastly, the case insensitive is just a convenience factor so i can be lazy.
            comms_return_char: character to use to send returns to host
            comms_ansi: True/False strip comms_ansi characters from output
            ssh_config_file: string to path for ssh config file, True to use default ssh config file
                or False to ignore default ssh config file
            ssh_known_hosts_file: string to path for ssh known hosts file, True to use default known
                file locations. Only applicable/needed if `auth_strict_key` is set to True
            on_open: callable that accepts the class instance as its only argument. this callable,
                if provided, is executed immediately after authentication is completed. Common use
                cases for this callable would be to disable paging or accept any kind of banner
                message that prompts a user upon connection
            on_close: callable that accepts the class instance as its only argument. this callable,
                if provided, is executed immediately prior to closing the underlying transport.
                Common use cases for this callable would be to save configurations prior to exiting,
                or to logout properly to free up vtys or similar.
            transport: system|telnet or a plugin -- type of transport to use for connection
                system uses system available ssh (/usr/bin/ssh)
                ssh2 uses ssh2-python *has been migrated to a plugin
                paramiko uses... paramiko *has been migrated to a plugin
                telnet uses telnetlib
                choice of driver depends on the features you need. in general system is easiest as
                it will just 'auto-magically' use your ssh config file ('~/.ssh/config' or
                '/etc/ssh/config_file'). ssh2 is very very fast as it is a thin wrapper around
                libssh2 however it is slightly feature limited. paramiko is slower than ssh2, but
                has more features built in (though scrapli does not expose/support them all).
            transport_options: dictionary of options to pass to selected transport class; see
                docs for given transport class for details of what to pass here

        Returns:
            N/A  # noqa: DAR202

        Raises:
            UnsupportedPlatform: if using windows with system transport

        """
        # create a dict of all "initialization" args for posterity and for passing to Transport
        # and Channel objects
        self._initialization_args: Dict[str, Any] = {}

        self._setup_host(host=host, port=port)
        self.logger: Logger = getLogger(f"scrapli.driver-{self._host}")

        self._setup_auth(
            auth_username=auth_username,
            auth_password=auth_password,
            auth_private_key=auth_private_key,
            auth_private_key_passphrase=auth_private_key_passphrase,
            auth_strict_key=auth_strict_key,
            auth_bypass=auth_bypass,
        )
        self._setup_timeouts(
            timeout_socket=timeout_socket,
            timeout_transport=timeout_transport,
            timeout_ops=timeout_ops,
            timeout_exit=timeout_exit,
        )
        self._timeout_ops: float = self._initialization_args["timeout_ops"]
        self._setup_keepalive(
            keepalive=keepalive,
            keepalive_type=keepalive_type,
            keepalive_interval=keepalive_interval,
            keepalive_pattern=keepalive_pattern,
        )
        self._setup_comms(
            comms_prompt_pattern=comms_prompt_pattern,
            comms_return_char=comms_return_char,
            comms_ansi=comms_ansi,
        )
        self._setup_callables(on_open=on_open, on_close=on_close)

        if transport not in ("system", "telnet"):
            self.logger.info(f"Non-core transport `{transport}` selected")
        self._transport = transport

        if transport == "system" and sys.platform.startswith("win"):
            msg = "`system` transport is not supported on Windows, please use a different transport"
            raise UnsupportedPlatform(msg)

        if transport != "telnet":
            self._setup_ssh_args(
                ssh_config_file=ssh_config_file, ssh_known_hosts_file=ssh_known_hosts_file
            )

        self._initialization_args["transport_options"] = transport_options
        self.transport_class, self.transport_args = self._transport_factory(transport=transport)
        # so transport drivers don't need to support `transport_options` as an argument, if no
        # transport options provided, do nothing, otherwise add this to the args we ship to the
        # transport class
        if transport_options is not None:
            self.transport_args["transport_options"] = transport_options
        self.transport = self.transport_class(**self.transport_args)

        self.channel_args: Dict[str, Any] = {}
        for arg in CHANNEL_ARGS:
            if arg == "transport":
                continue
            self.channel_args[arg] = self._initialization_args.get(arg)

    def __str__(self) -> str:
        """
        Magic str method for Scrape

        Args:
            N/A

        Returns:
            str: str representation of object

        Raises:
            N/A

        """
        return f"Scrape Object for host {self._host}"

    def __repr__(self) -> str:
        """
        Magic repr method for Scrape

        Args:
            N/A

        Returns:
            str: repr for class object

        Raises:
            N/A

        """
        return (
            f"{self.__class__.__name__}("
            f"host={self._initialization_args['host']!r}, "
            f"port={self._initialization_args['port']!r}, "
            f"auth_username={self._initialization_args['auth_username']!r}, "
            f"auth_password={self._initialization_args['auth_password']!r}, "
            f"auth_private_key={self._initialization_args['auth_private_key']!r}, "
            "auth_private_key_passphrase="
            f"{self._initialization_args['auth_private_key_passphrase']!r}, "
            f"auth_strict_key={self._initialization_args['auth_strict_key']!r}, "
            f"auth_bypass={self._initialization_args['auth_bypass']!r}, "
            f"timeout_socket={self._initialization_args['timeout_socket']!r}, "
            f"timeout_transport={self._initialization_args['timeout_transport']!r}, "
            f"timeout_ops={self._initialization_args['timeout_ops']!r}, "
            f"timeout_exit={self._initialization_args['timeout_exit']!r}, "
            f"keepalive={self._initialization_args['keepalive']!r}, "
            f"keepalive_interval={self._initialization_args['keepalive_interval']!r}, "
            f"keepalive_type={self._initialization_args['keepalive_type']!r}, "
            f"keepalive_pattern={self._initialization_args['keepalive_pattern']!r}, "
            f"comms_prompt_pattern={self._initialization_args['comms_prompt_pattern']!r}, "
            f"comms_return_char={self._initialization_args['comms_return_char']!r}, "
            f"comms_ansi={self._initialization_args['comms_ansi']!r}, "
            f"ssh_config_file={self._initialization_args.get('ssh_config_file')!r}, "
            f"ssh_known_hosts_file={self._initialization_args.get('ssh_known_hosts_file')!r}, "
            f"on_open={self.on_open!r}, "
            f"on_close={self.on_close!r}, "
            f"transport={self._transport!r}, "
            f"transport_options={self._initialization_args.get('transport_options')!r})"
        )

    def _setup_host(self, host: str, port: int) -> None:
        """
        Parse and setup host attributes

        Args:
            host: host to parse/set
            port: port to parse/set

        Returns:
            N/A  # noqa: DAR202

        Raises:
            ValueError: if host is not provided
            TypeError: if port is not an integer

        """
        if not host:
            raise ValueError("`host` should be a hostname/ip address, got nothing!")
        if not isinstance(port, int):
            raise TypeError(f"`port` should be int, got {type(port)}")
        self._host = host.strip()
        self._initialization_args["host"] = host.strip()
        self._initialization_args["port"] = port

    def _setup_auth(
        self,
        auth_username: str,
        auth_password: str,
        auth_private_key: str,
        auth_private_key_passphrase: str,
        auth_strict_key: bool,
        auth_bypass: bool,
    ) -> None:
        """
        Parse and setup auth attributes

        Args:
            auth_username: username to parse/set
            auth_password: password to parse/set
            auth_private_key: ssh key to parse/set
            auth_private_key_passphrase: ssh key passphrase to parse/set
            auth_strict_key: strict key to parse/set
            auth_bypass: bypass to parse/set

        Returns:
            N/A  # noqa: DAR202

        Raises:
            TypeError: if auth_strict_key is not a bool
            ValueError: if auth_private_key is not a valid file

        """
        if not isinstance(auth_strict_key, bool):
            raise TypeError(f"`auth_strict_key` should be bool, got {type(auth_strict_key)}")
        if not isinstance(auth_bypass, bool):
            raise TypeError(f"`auth_bypass` should be bool, got {type(auth_bypass)}")

        self._initialization_args["auth_strict_key"] = auth_strict_key
        self._initialization_args["auth_bypass"] = auth_bypass
        self._initialization_args["auth_username"] = auth_username.strip()
        self._initialization_args["auth_password"] = auth_password.strip()
        self._initialization_args[
            "auth_private_key_passphrase"
        ] = auth_private_key_passphrase.strip()

        if auth_private_key:
            private_key_path = Path.expanduser(Path(auth_private_key.strip()))
            if not private_key_path.is_file():
                raise ValueError(f"Provided public key `{auth_private_key}` is not a file")
            self._initialization_args["auth_private_key"] = os.path.expanduser(
                auth_private_key.strip()
            )
        else:
            self._initialization_args["auth_private_key"] = auth_private_key

    def _setup_timeouts(
        self,
        timeout_socket: int,
        timeout_transport: int,
        timeout_ops: float,
        timeout_exit: bool,
    ) -> None:
        """
        Parse and setup timeout attributes

        Args:
            timeout_socket: socket timeout to parse/set
            timeout_transport: transport timeout to parse/set
            timeout_ops: ops timeout to parse/set
            timeout_exit: timeout exit bool to parse/set

        Returns:
            N/A  # noqa: DAR202

        Raises:
            TypeError: if invalid type args provided

        """
        if not isinstance(timeout_exit, bool):
            raise TypeError(f"`timeout_exit` should be bool, got {type(timeout_exit)}")

        self._initialization_args["timeout_socket"] = int(timeout_socket)
        self._initialization_args["timeout_transport"] = int(timeout_transport)
        self._initialization_args["timeout_ops"] = float(timeout_ops)
        self._initialization_args["timeout_exit"] = timeout_exit

    def _setup_keepalive(
        self, keepalive: bool, keepalive_type: str, keepalive_interval: int, keepalive_pattern: str
    ) -> None:
        """
        Parse and setup keepalive attributes

        Args:
            keepalive: keepalive to parse/set
            keepalive_type: keepalive_type to parse/set
            keepalive_interval: keepalive_interval to parse/set
            keepalive_pattern: keepalive_pattern to parse/set

        Returns:
            N/A  # noqa: DAR202

        Raises:
            TypeError: if keepalive is not a bool
            ValueError: if keepalive_type is not valid

        """
        if not isinstance(keepalive, bool):
            raise TypeError(f"`keepalive` should be bool, got {type(keepalive)}")
        if keepalive_type not in ["network", "standard"]:
            raise ValueError(
                f"`{keepalive_type}` is an invalid keepalive_type; must be 'network' or 'standard'"
            )
        self._initialization_args["keepalive"] = keepalive
        self._initialization_args["keepalive_interval"] = int(keepalive_interval)
        self._initialization_args["keepalive_type"] = keepalive_type
        self._initialization_args["keepalive_pattern"] = keepalive_pattern

    def _setup_comms(
        self, comms_prompt_pattern: str, comms_return_char: str, comms_ansi: bool
    ) -> None:
        """
        Parse and setup auth attributes

        Args:
            comms_prompt_pattern: prompt pattern to parse/set
            comms_return_char: return char to parse/set
            comms_ansi: ansi val to parse/set

        Returns:
            N/A  # noqa: DAR202

        Raises:
            TypeError: if invalid type args provided

        """
        # try to compile prompt to raise TypeError before opening any connections
        re.compile(comms_prompt_pattern, flags=re.M | re.I)

        if not isinstance(comms_return_char, str):
            raise TypeError(f"`comms_return_char` should be str, got {type(comms_return_char)}")
        if not isinstance(comms_ansi, bool):
            raise TypeError(f"`comms_ansi` should be bool, got {type(comms_ansi)}")

        self._initialization_args["comms_prompt_pattern"] = comms_prompt_pattern
        self._initialization_args["comms_return_char"] = comms_return_char
        self._initialization_args["comms_ansi"] = comms_ansi

    def _setup_callables(
        self, on_open: Optional[Callable[..., Any]], on_close: Optional[Callable[..., Any]]
    ) -> None:
        """
        Parse and setup callables (on_open/on_close)

        Args:
            on_open: on_open to parse/set
            on_close: on_close to parse/set

        Returns:
            N/A  # noqa: DAR202

        Raises:
            TypeError: if port is not an integer

        """
        if on_open is not None and not callable(on_open):
            raise TypeError(f"`on_open` must be a callable, got {type(on_open)}")
        if on_close is not None and not callable(on_close):
            raise TypeError(f"`on_close` must be a callable, got {type(on_close)}")
        self.on_open = on_open
        self.on_close = on_close
        self._initialization_args["on_open"] = on_open
        self._initialization_args["on_close"] = on_close

    def _setup_ssh_args(
        self, ssh_config_file: Union[str, bool], ssh_known_hosts_file: Union[str, bool]
    ) -> None:
        """
        Parse and setup ssh related arguments

        Args:
            ssh_config_file: string to path for ssh config file, True to use default ssh config file
                or False to ignore default ssh config file
            ssh_known_hosts_file: string to path for ssh known hosts file, True to use default known
                file locations. Only applicable/needed if `auth_strict_key` is set to True

        Returns:
            N/A  # noqa: DAR202

        Raises:
            TypeError: if invalid config file or known hosts file value provided

        """
        if not isinstance(ssh_config_file, (str, bool)):
            raise TypeError(f"`ssh_config_file` must be str or bool, got {type(ssh_config_file)}")
        if not isinstance(ssh_known_hosts_file, (str, bool)):
            raise TypeError(
                "`ssh_known_hosts_file` must be str or bool, got " f"{type(ssh_known_hosts_file)}"
            )

        if ssh_config_file is not False:
            if isinstance(ssh_config_file, bool):
                cfg = ""
            else:
                cfg = ssh_config_file
            self._initialization_args["ssh_config_file"] = resolve_ssh_config(cfg)
        else:
            self._initialization_args["ssh_config_file"] = ""

        if ssh_known_hosts_file is not False:
            if isinstance(ssh_known_hosts_file, bool):
                known_hosts = ""
            else:
                known_hosts = ssh_known_hosts_file
            self._initialization_args["ssh_known_hosts_file"] = resolve_ssh_known_hosts(known_hosts)
        else:
            self._initialization_args["ssh_known_hosts_file"] = ""

    def _transport_factory(self, transport: str) -> Tuple[Callable[..., Any], Dict[str, Any]]:
        """
        Private factory method to produce transport class

        Args:
            transport: string name of transport class to use

        Returns:
            Transport: initialized Transport class

        Raises:
            N/A  # noqa

        """
        transport_class = TRANSPORT_CLASS.get(transport, None)
        required_transport_args = TRANSPORT_ARGS.get(transport, ())

        if transport_class is None:
            transport_class, required_transport_args = _find_transport_plugin(transport=transport)

        transport_args = {}
        for arg in TRANSPORT_BASE_ARGS:
            transport_args[arg] = self._initialization_args.get(arg)
        for arg in required_transport_args:
            transport_args[arg] = self._initialization_args.get(arg)
        return transport_class, transport_args

    def isalive(self) -> bool:
        """
        Check if underlying socket/channel is alive

        Args:
            N/A

        Returns:
            bool: True/False if socket/channel is alive

        Raises:
            N/A

        """
        alive = False
        try:
            alive = self.transport.isalive()
        except AttributeError:
            pass
        return alive



"""mock_devices.cisco_iosxe"""
from typing import Any

from .base import BaseServer, BaseSSHServerSession

SHOW_COMMAND = """Cisco IOS XE Software, Version 16.04.01
Cisco IOS Software [Everest], CSR1000V Software (X86_64_LINUX_IOSD-UNIVERSALK9-M), Version 16.4.1, RELEASE SOFTWARE (fc2)
Technical Support: http://www.cisco.com/techsupport
Copyright (c) 1986-2016 by Cisco Systems, Inc.
Compiled Sun 27-Nov-16 13:02 by mcpre


Cisco IOS-XE software, Copyright (c) 2005-2016 by cisco Systems, Inc.
All rights reserved.  Certain components of Cisco IOS-XE software are
licensed under the GNU General Public License ("GPL") Version 2.0.  The
software code licensed under GPL Version 2.0 is free software that comes
with ABSOLUTELY NO WARRANTY.  You can redistribute and/or modify such
GPL code under the terms of GPL Version 2.0.  For more details, see the
documentation or "License Notice" file accompanying the IOS-XE software,
or the applicable URL provided on the flyer accompanying the IOS-XE
software.


ROM: IOS-XE ROMMON

csr1000v uptime is 4 hours, 55 minutes
Uptime for this control processor is 4 hours, 56 minutes
System returned to ROM by reload
System image file is "bootflash:packages.conf"
Last reload reason: reload



This product contains cryptographic features and is subject to United
States and local country laws governing import, export, transfer and
use. Delivery of Cisco cryptographic products does not imply
third-party authority to import, export, distribute or use encryption.
Importers, exporters, distributors and users are responsible for
compliance with U.S. and local country laws. By using this product you
agree to comply with applicable laws and regulations. If you are unable
to comply with U.S. and local laws, return this product immediately.

A summary of U.S. laws governing Cisco cryptographic products may be found at:
http://www.cisco.com/wwl/export/crypto/tool/stqrg.html

If you require further assistance please contact us by sending email to
export@cisco.com.

License Level: ax
License Type: Default. No valid license found.
Next reload license Level: ax

cisco CSR1000V (VXE) processor (revision VXE) with 2052375K/3075K bytes of memory.
Processor board ID 9FKLJWM5EB0
10 Gigabit Ethernet interfaces
32768K bytes of non-volatile configuration memory.
3985132K bytes of physical memory.
7774207K bytes of virtual hard disk at bootflash:.
0K bytes of  at webui:.

Configuration register is 0x2102
"""  # noqa: E0501


class IOSXESSHServerSession(BaseSSHServerSession):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Cisco IOSXE mock ssh server

        Args:
            args: args to pass to base class
            kwargs: keyword args to pass to base class

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        super().__init__(*args, **kwargs)
        self.priv_levels = {
            "exec": "csr1000v>",
            "privilege_exec": "csr1000v#",
            "config": "csr1000v(config)#",
        }
        self.priv_level = "privilege_exec"
        self.invalid_input = "% Invalid input detected at '^' marker.\n"
        self.paging_prompt = "\n--More--"

        self.command_mapper = {
            self.handle_open: ["terminal length 0", "terminal width 0"],
            self.handle_close: ["exit"],
            self.handle_priv: ["disable", "enable", "configure terminal", "end"],
            self.handle_show: ["show version"],
            self.handle_interactive: ["clear logging"],
        }

        self.show_command_output = SHOW_COMMAND

    def handle_open(self, channel_input: str) -> None:
        """
        Handle "open" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if self.priv_level == "config":
            self.handle_invalid(channel_input=channel_input)
        elif channel_input == "terminal length 0":
            # in the future can actually set the value if we care?
            self.terminal_length = 0
            self._chan.write("")
            self.repaint_prompt()
        elif channel_input == "terminal width 0":
            self.terminal_width = 0
            self._chan.write("")
            self.repaint_prompt()
        else:
            self.handle_invalid(channel_input=channel_input)

    def handle_close(self, channel_input: str) -> None:
        """
        Handle "close" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        _ = channel_input
        if self.priv_level == "config":
            # exit is a weird situation since it can be used to exit config mode or exit the device
            # so this lives here instead of handle_priv
            self._chan.write("")
            self.priv_level = "privilege_exec"
            self.repaint_prompt()
        else:
            self.eof_received()

    def _handle_priv_password_input(self, channel_input: str) -> None:
        """
        Handle the enable password input

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if channel_input == "scrapli":
            self._chan.set_echo(True)
            self._hide_input = False
            self.priv_level = "privilege_exec"
            self._chan.write("")
            self._chan.write(self.priv_levels.get(self.priv_level))
            self._redirect_to_handler = None
        else:
            self._chan.write("Password:")
            return

    def _handle_priv_disable(self, channel_input: str) -> None:
        """
        Handle the "disable" command

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if self.priv_level == "privilege_exec":
            self._chan.write("")
            self.priv_level = "exec"
            self.repaint_prompt()
        elif self.priv_level == "exec":
            self._chan.write("")
            self.repaint_prompt()
        else:
            self.handle_invalid(channel_input=channel_input)

    def _handle_priv_enable(self, channel_input: str) -> None:
        """
        Handle the "enable" command

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if self.priv_level == "exec":
            if self._redirect_to_handler is None:
                self._chan.set_echo(False)
                self._hide_input = True
                self._chan.write("Password:")
                self._redirect_to_handler = self._handle_priv_password_input
        else:
            self.handle_invalid(channel_input=channel_input)

    def _handle_priv_configure(self, channel_input: str) -> None:
        """
        Handle the "configure terminal" command

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if self.priv_level != "privilege_exec":
            self.handle_invalid(channel_input=channel_input)
        else:
            self._chan.write("")
            self.priv_level = "config"
            self.repaint_prompt()

    def _handle_priv_end(self, channel_input: str) -> None:
        """
        Handle the "end" command

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if self.priv_level != "config":
            self.handle_invalid(channel_input=channel_input)
        else:
            self._chan.write("")
            self.priv_level = "privilege_exec"
            self.repaint_prompt()

    def handle_priv(self, channel_input: str) -> None:
        """
        Handle "privilege" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if channel_input == "disable":
            self._handle_priv_disable(channel_input=channel_input)
        elif channel_input == "enable":
            self._handle_priv_enable(channel_input=channel_input)
        elif channel_input == "configure terminal":
            self._handle_priv_configure(channel_input=channel_input)
        elif channel_input == "end":
            self._handle_priv_end(channel_input=channel_input)

    def handle_interactive(self, channel_input: str) -> None:
        """
        Handle the "interactive" (clear logging) command

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if self.priv_level == "privilege_exec":
            if self._redirect_to_handler is None:
                self._chan.write("Clear logging buffer [confirm]")
        else:
            self.handle_invalid(channel_input=channel_input)

    def data_received(self, data: str, datatype: None) -> None:
        """
        Handle data received on ssh channel

        Args:
            data: string of data sent to channel
            datatype: dunno! in base class though :)

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        _ = datatype
        # in the future we can cutoff the inputs if it is over X width if disable width has not yet
        # been ran -- not needed now but could be cool; if we just send a return, we should NOT
        # strip that!
        channel_input = data if data == "\n" else data.rstrip()

        if self._redirect_to_handler is not None:
            # "redirect" to a function that needs continued inputs such as priv escalation
            self._redirect_to_handler(channel_input=channel_input)
            return

        if data == "\n":
            self.repaint_prompt()
            return

        for handler, channel_inputs in self.command_mapper.items():
            if channel_input in channel_inputs:
                # as soon as we find a match for the channel input ship it to the handler; handler
                # can figure out if command is valid for the given priv and such
                handler(channel_input=channel_input)
                return

        self.handle_invalid(channel_input=channel_input)


class IOSXEServer(BaseServer):
    pass



from pathlib import Path

import pytest

import scrapli
from scrapli.channel import Channel
from scrapli.transport import Transport

from ...test_data.unit_test_cases import TEST_CASES

TEST_DATA_DIR = f"{Path(scrapli.__file__).parents[1]}/tests/test_data"


def test_init(sync_generic_driver_conn):
    """Test that all arguments get properly passed from driver/transport to channel on init"""
    assert isinstance(sync_generic_driver_conn.channel.transport, Transport)
    assert isinstance(sync_generic_driver_conn.channel, Channel)


@pytest.mark.parametrize(
    "strip_prompt",
    [True, False],
    ids=["strip_prompt", "no_strip_prompt"],
)
def test_send_command(sync_generic_driver_conn, strip_prompt):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_input"]["raw_result"]
    expected_processed = (
        TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["strip"]
        if strip_prompt
        else TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["no_strip"]
    )
    sync_generic_driver_conn.open()
    sync_generic_driver_conn.send_command(command="terminal length 0")
    response = sync_generic_driver_conn.send_command(
        command="show version", strip_prompt=strip_prompt
    )
    assert response.raw_result == expected_raw.encode()
    assert response.result == expected_processed


@pytest.mark.parametrize(
    "strip_prompt",
    [True, False],
    ids=["strip_prompt", "no_strip_prompt"],
)
def test_send_commands(sync_generic_driver_conn, strip_prompt):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_input"]["raw_result"]
    expected_processed = (
        TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["strip"]
        if strip_prompt
        else TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["no_strip"]
    )
    sync_generic_driver_conn.open()
    multi_response = sync_generic_driver_conn.send_commands(
        commands=["terminal length 0", "show version"], strip_prompt=strip_prompt
    )
    assert multi_response[1].raw_result == expected_raw.encode()
    assert multi_response[1].result == expected_processed


@pytest.mark.parametrize(
    "strip_prompt",
    [True, False],
    ids=["strip_prompt", "no_strip_prompt"],
)
def test_send_commands_from_file(sync_generic_driver_conn, strip_prompt):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_input"]["raw_result"]
    expected_processed = (
        TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["strip"]
        if strip_prompt
        else TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["no_strip"]
    )
    sync_generic_driver_conn.open()
    sync_generic_driver_conn.send_command(command="terminal length 0")
    multi_response = sync_generic_driver_conn.send_commands_from_file(
        file=f"{TEST_DATA_DIR}/files/cisco_iosxe_commands", strip_prompt=strip_prompt
    )
    assert multi_response[0].raw_result == expected_raw.encode()
    assert multi_response[0].result == expected_processed


def test_send_interactive(sync_generic_driver_conn):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_inputs_interact"]["raw_result"]
    expected_processed = TEST_CASES["cisco_iosxe"]["test_send_inputs_interact"]["processed_result"]
    interact_events = TEST_CASES["cisco_iosxe"]["test_send_inputs_interact"]["interact_events"]
    sync_generic_driver_conn.open()
    response = sync_generic_driver_conn.send_interactive(interact_events=interact_events)
    assert expected_raw.encode() in response.raw_result
    assert expected_processed in response.result


def test_get_prompt(sync_generic_driver_conn):
    expected_prompt = TEST_CASES["cisco_iosxe"]["test_get_prompt"]["privilege_exec"]
    sync_generic_driver_conn.open()
    found_prompt = sync_generic_driver_conn.get_prompt()
    assert found_prompt == expected_prompt



import logging
from logging import LogRecord


def test_duplicate_filter():
    logger = logging.getLogger("scrapli")
    log_entry = LogRecord(
        name="scrapli_test",
        level="DEBUG",
        pathname="somepath",
        lineno=123,
        args=None,
        exc_info=(),
        msg="Fake as heck",
    )
    logger.last_log = log_entry
    assert logger.filter(record=log_entry) is True
    assert logger.filter(record=log_entry) is False



import logging
import time
from pathlib import Path

from device_info import device

from scrapli.driver.core import IOSXEDriver

logging.basicConfig(
    filename=f"{Path(__file__).resolve().parents[0]}/ssh_config_file_and_key_auth.log",
    level=logging.DEBUG,
)
logger = logging.getLogger("scrapli")

args = {
    "host": device["host"],
    "port": device["port"],
    "ssh_config_file": True,
    "auth_strict_key": False,
    "keepalive_interval": device["keepalive_interval"],
    "transport": device["transport"],
    "keepalive": device["keepalive"],
}

conn = IOSXEDriver(**args)
conn.open()

print("***** Get Prompt:")
print(conn.get_prompt())

print("***** Show run | i hostname:")
result = conn.send_commands("show run | i hostname")
print(result, result[0].result)

print("***** Clear logging buffer:")
interact = [("clear logg", "Clear logging buffer [confirm]"), ("", "3560CX#")]
result = conn.send_interactive(interact)
print(result, result[0].result)

print("***** Disable Paging:")
result = conn.send_commands("term length 0")
print(result, result[0].result)

print("***** Show run:")
result = conn.send_commands("show run")
print(result, result[0].result)

if device["keepalive"]:
    print("***** Waiting for keepalive....")
    time.sleep(5)

conn.close()



"""scrapli.transport.telnet"""
import re
from datetime import datetime
from select import select
from telnetlib import Telnet
from typing import Optional

from scrapli.decorators import OperationTimeout, requires_open_session
from scrapli.exceptions import ScrapliAuthenticationFailed
from scrapli.helper import get_prompt_pattern, strip_ansi
from scrapli.transport.transport import Transport

TELNET_TRANSPORT_ARGS = (
    "auth_username",
    "auth_password",
    "auth_bypass",
    "comms_prompt_pattern",
    "comms_return_char",
    "comms_ansi",
    "timeout_ops",
)


class ScrapliTelnet(Telnet):
    def __init__(self, host: str, port: int, timeout: int) -> None:
        """
        ScrapliTelnet class for typing purposes

        Args:
            host: string of host
            port: integer port to connect to
            timeout: timeout value in seconds

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.eof: bool
        self.timeout: int
        super().__init__(host, port, timeout)


class TelnetTransport(Transport):
    def __init__(
        self,
        host: str,
        port: int = 23,
        auth_username: str = "",
        auth_password: str = "",
        auth_bypass: bool = False,
        timeout_socket: int = 5,
        timeout_transport: int = 5,
        timeout_ops: int = 10,
        timeout_exit: bool = True,
        keepalive: bool = False,
        keepalive_interval: int = 30,
        keepalive_type: str = "",
        keepalive_pattern: str = "\005",
        comms_prompt_pattern: str = r"^[a-z0-9.\-@()/:]{1,32}[#>$]$",
        comms_return_char: str = "\n",
        comms_ansi: bool = False,
    ) -> None:
        r"""
        TelnetTransport Object

        Inherit from Transport ABC
        TelnetTransport <- Transport (ABC)

        Note that comms_prompt_pattern, comms_return_char and comms_ansi are only passed here to
        handle "in channel" authentication required by SystemSSH -- these are assigned to private
        attributes in this class and ignored after authentication. If you wish to modify these
        values on a "live" scrapli connection, modify them in the Channel object, i.e.
        `conn.channel.comms_prompt_pattern`. Additionally timeout_ops is passed and assigned to
        _timeout_ops to use the same timeout_ops that is used in Channel to decorate the
        authentication methods here.

        Args:
            host: host ip/name to connect to
            port: port to connect to
            auth_username: username for authentication
            auth_password: password for authentication
            auth_bypass: bypass authentication process
            timeout_socket: timeout for establishing socket in seconds -- since this is not directly
                exposed in telnetlib, this is just the initial timeout for the telnet connection.
                After the connection is established, the timeout is modified to the value of
                `timeout_transport`.
            timeout_transport: timeout for telnet transport in seconds
            timeout_ops: timeout for telnet channel operations in seconds -- this is also the
                timeout for finding and responding to username and password prompts at initial
                login. This is assigned to a private attribute and is ignored after authentication
                is completed.
            timeout_exit: True/False close transport if timeout encountered. If False and keepalives
                are in use, keepalives will prevent program from exiting so you should be sure to
                catch Timeout exceptions and handle them appropriately
            keepalive: whether or not to try to keep session alive
            keepalive_interval: interval to use for session keepalives
            keepalive_type: network|standard -- 'network' sends actual characters over the
                transport channel. This is useful for network-y type devices that may not support
                "standard" keepalive mechanisms. 'standard' is not currently implemented for telnet
            keepalive_pattern: pattern to send to keep network channel alive. Default is
                u'\005' which is equivalent to 'ctrl+e'. This pattern moves cursor to end of the
                line which should be an innocuous pattern. This will only be entered *if* a lock
                can be acquired. This is only applicable if using keepalives and if the keepalive
                type is 'network'
            comms_prompt_pattern: prompt pattern expected for device, same as the one provided to
                channel -- telnet needs to know this to know how to decide if we are properly
                sending/receiving data -- i.e. we are not stuck at some password prompt or some
                other failure scenario. If using driver, this should be passed from driver (Scrape,
                or IOSXE, etc.) to this Transport class. This is assigned to a private attribute and
                is ignored after authentication is completed.
            comms_return_char: return character to use on the channel, same as the one provided to
                channel -- telnet needs to know this to know what to send so that we can probe
                the channel to make sure we are authenticated and sending/receiving data. If using
                driver, this should be passed from driver (Scrape, or IOSXE, etc.) to this Transport
                class. This is assigned to a private attribute and is ignored after authentication
                is completed.
            comms_ansi: True/False strip comms_ansi characters from output; this value is assigned
                self._comms_ansi and is ignored after authentication. We only need it for transport
                on the off chance (maybe never, especially here in telnet land?) that
                username/password prompts contain ansi characters, otherwise "comms_ansi" is really
                a channel attribute and is treated as such. This is assigned to a private attribute
                and is ignored after authentication is completed.

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        super().__init__(
            host,
            port,
            timeout_socket,
            timeout_transport,
            timeout_exit,
            keepalive,
            keepalive_interval,
            keepalive_type,
            keepalive_pattern,
        )
        self.auth_username: str = auth_username
        self.auth_password: str = auth_password
        self.auth_bypass: bool = auth_bypass

        self._timeout_ops: int = timeout_ops
        # timeout_ops_auth is only used for authentication; base ops timeout * 2 as we are doing
        # two operations -- entering username and entering password (in most cases at least)
        self._timeout_ops_auth: int = timeout_ops * 2

        self._comms_prompt_pattern: str = comms_prompt_pattern
        self._comms_return_char: str = comms_return_char
        self._comms_ansi: bool = comms_ansi

        self.username_prompt: str = "Username:"
        self.password_prompt: str = "Password:"

        self.session: ScrapliTelnet
        self.lib_auth_exception = ScrapliAuthenticationFailed
        self._isauthenticated = False

    def open(self) -> None:
        """
        Open telnet channel

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            ScrapliAuthenticationFailed: if cant successfully authenticate

        """
        self.session_lock.acquire()
        # establish session with "socket" timeout, then reset timeout to "transport" timeout
        try:
            telnet_session = ScrapliTelnet(
                host=self.host, port=self.port, timeout=self.timeout_socket
            )
        except ConnectionError as exc:
            msg = f"Failed to open telnet session to host {self.host}"
            if "connection refused" in str(exc).lower():
                msg = f"Failed to open telnet session to host {self.host}, connection refused"
            raise ScrapliAuthenticationFailed(msg) from exc
        telnet_session.timeout = self.timeout_transport
        self.logger.debug(f"Session to host {self.host} spawned")
        self.session_lock.release()

        if not self.auth_bypass:
            self._authenticate(telnet_session=telnet_session)
        else:
            self.logger.info("`auth_bypass` is True, bypassing authentication")
            # if we skip auth, we'll manually set _isauthenticated to True
            self._isauthenticated = True

        if not self._isauthenticated and not self._telnet_isauthenticated(
            telnet_session=telnet_session
        ):
            raise ScrapliAuthenticationFailed(
                f"Could not authenticate over telnet to host: {self.host}"
            )

        self.logger.debug(f"Authenticated to host {self.host} with password")
        self.session = telnet_session

    @OperationTimeout("_timeout_ops_auth", "Timed out looking for telnet login prompts")
    def _authenticate(self, telnet_session: ScrapliTelnet) -> None:
        """
        Parent private method to handle telnet authentication

        Args:
            telnet_session: Telnet session object

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.session_lock.acquire()
        output = b""

        # capture the start time of the authentication event; we also set a "return_interval" which
        # is 1/10 the timout_ops value, we will send a return character at roughly this interval if
        # there is no output on the channel. we do this because sometimes telnet needs a kick to get
        # it to prompt for auth -- particularity when connecting to terminal server/console port
        auth_start_time = datetime.now().timestamp()
        return_interval = self._timeout_ops / 10
        return_attempts = 1

        while True:
            output += telnet_session.read_eager()
            if self.username_prompt.lower().encode() in output.lower():
                # if/when we see username, reset the output to empty byte string
                output = b""
                telnet_session.write(self.auth_username.encode())
                telnet_session.write(self._comms_return_char.encode())
            elif self.password_prompt.lower().encode() in output.lower():
                telnet_session.write(self.auth_password.encode())
                telnet_session.write(self._comms_return_char.encode())
                self.session_lock.release()
                break
            elif not output:
                current_iteration_time = datetime.now().timestamp()
                if (current_iteration_time - auth_start_time) > (return_interval * return_attempts):
                    telnet_session.write(self._comms_return_char.encode())
                    return_attempts += 1

    @OperationTimeout("_timeout_ops_auth", "Timed determining if telnet session is authenticated")
    def _telnet_isauthenticated(self, telnet_session: ScrapliTelnet) -> bool:
        """
        Check if session is authenticated

        This is very naive -- it only knows if the sub process is alive and has not received an EOF.
        Beyond that we lock the session and send the return character and re-read the channel.

        Args:
            telnet_session: Telnet session object

        Returns:
            bool: True if authenticated, else False

        Raises:
            N/A

        """
        self.logger.debug("Attempting to determine if telnet authentication was successful")
        if not telnet_session.eof:
            prompt_pattern = get_prompt_pattern(prompt="", class_prompt=self._comms_prompt_pattern)
            telnet_session_fd = telnet_session.fileno()
            self.session_lock.acquire()
            telnet_session.write(self._comms_return_char.encode())
            while True:
                # almost all of the time we don't need a while loop here, but every once in a while
                # fd won't be ready which causes a failure without an obvious root cause,
                # loop/logging to hopefully help with that
                fd_ready, _, _ = select([telnet_session_fd], [], [], 0)
                if telnet_session_fd in fd_ready:
                    break
                self.logger.debug("PTY fd not ready yet...")
            output = b""
            while True:
                output += telnet_session.read_eager()
                # we do not need to deal w/ line replacement for the actual output, only for
                # parsing if a prompt-like thing is at the end of the output
                output = output.replace(b"\r", b"")
                if self._comms_ansi:
                    output = strip_ansi(output=output)
                channel_match = re.search(pattern=prompt_pattern, string=output)
                if channel_match:
                    self.session_lock.release()
                    self._isauthenticated = True
                    return True
                if b"password:" in output.lower():
                    # if we see "password" auth failed... hopefully true in all scenarios!
                    self.session_lock.release()
                    return False
        self.session_lock.release()
        return False

    def close(self) -> None:
        """
        Close session and socket

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.session_lock.acquire()
        self.session.close()
        self.logger.debug(f"Channel to host {self.host} closed")
        self.session_lock.release()

    def isalive(self) -> bool:
        """
        Check if alive and session is authenticated

        Args:
            N/A

        Returns:
            bool: True if alive, False otherwise.

        Raises:
            N/A

        """
        if self._isauthenticated and not self.session.eof:
            return True
        return False

    @requires_open_session()
    @OperationTimeout("timeout_transport", "Timed out reading from transport")
    def read(self) -> bytes:
        """
        Read data from the channel

        Args:
            N/A

        Returns:
            bytes: bytes read from the telnet channel

        Raises:
            N/A

        """
        return self.session.read_eager()

    @requires_open_session()
    @OperationTimeout("timeout_transport", "Timed out writing to transport")
    def write(self, channel_input: str) -> None:
        """
        Write data to the channel

        Args:
            channel_input: string to send to channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.session.write(channel_input.encode())

    @requires_open_session()
    def set_timeout(self, timeout: Optional[int] = None) -> None:
        """
        Set session timeout

        Args:
            timeout: timeout in seconds

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if isinstance(timeout, int):
            set_timeout = timeout
        else:
            set_timeout = self.timeout_transport
        self.session.timeout = set_timeout

    def _keepalive_standard(self) -> None:
        """
        Send "out of band" (protocol level) keepalives to devices.

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            NotImplementedError: always, because this is not implemented for telnet

        """
        raise NotImplementedError("No 'standard' keepalive mechanism for telnet.")



from scrapli.driver.core import JunosDriver


def test_junos_driver_init_telnet():
    conn = JunosDriver(host="myhost", transport="telnet")
    assert conn.transport.username_prompt == "login:"


def test_junos_open_close(sync_juniper_junos_conn):
    sync_juniper_junos_conn.open()
    assert sync_juniper_junos_conn.get_prompt() == "vrnetlab>"
    # chalking this up to weirdness w/ the mock ssh server... it seems to not set the terminal
    # length correct on open the first time... unclear why... future me problem
    sync_juniper_junos_conn.send_command(command="set cli screen-length 0")
    # "more" will show up if we haven't sent terminal length 0 to the mock nxos server
    assert "more" not in sync_juniper_junos_conn.send_command(command="show version").result.lower()
    sync_juniper_junos_conn.close()



import os

host = os.getenv("SCRAPLI_SMOKE_HOST", None)
port = os.getenv("SCRAPLI_SMOKE_PORT", None)
user = os.getenv("SCRAPLI_SMOKE_USER", None)
password = os.getenv("SCRAPLI_SMOKE_PASS", None)

iosxe_device = {
    "host": host or "172.18.0.11",
    "port": port or 22,
    "auth_username": user or "vrnetlab",
    "auth_password": password or "VR-netlab9",
    "auth_strict_key": False,
    "transport": "system",
    "keepalive": True,
    "keepalive_interval": 1,
}

nxos_device = {
    "host": "172.18.0.12",
    "port": 22,
    "auth_username": "vrnetlab",
    "auth_password": "VR-netlab9",
    "auth_strict_key": False,
    "transport": "system",
    "keepalive": True,
    "keepalive_interval": 1,
}

iosxr_device = {
    "host": "172.18.0.13",
    "port": 22,
    "auth_username": "vrnetlab",
    "auth_password": "VR-netlab9",
    "auth_strict_key": False,
    "transport": "system",
    "keepalive": True,
    "keepalive_interval": 1,
}

eos_device = {
    "host": "172.18.0.14",
    "port": 22,
    "auth_username": "vrnetlab",
    "auth_password": "VR-netlab9",
    "auth_strict_key": False,
    "transport": "system",
    "keepalive": True,
    "keepalive_interval": 1,
    "comms_ansi": True,
}

junos_device = {
    "host": "172.18.0.15",
    "port": 22,
    "auth_username": "vrnetlab",
    "auth_password": "VR-netlab9",
    "auth_strict_key": False,
    "transport": "system",
    "keepalive": True,
    "keepalive_interval": 1,
}



from pathlib import Path

import scrapli

TEST_DATA_PATH = f"{Path(scrapli.__file__).parents[1]}/tests/test_data"

FUNCTIONAL_USERNAME = "vrnetlab"
FUNCTIONAL_PASSWORD = "VR-netlab9"
FUNCTIONAL_PASSPHRASE = "scrapli"
PRIVATE_KEY = f"{TEST_DATA_PATH}/files/vrnetlab_key"
ENCRYPTED_PRIVATE_KEY = f"{TEST_DATA_PATH}/files/vrnetlab_key_encrypted"
INVALID_PRIVATE_KEY = f"{TEST_DATA_PATH}/files/invalid_key"
MOCK_USERNAME = "scrapli"
MOCK_PASSWORD = "scrapli"
MOCK_PASSPHRASE = FUNCTIONAL_PASSPHRASE

DEVICES = {
    "cisco_iosxe": {
        "auth_username": FUNCTIONAL_USERNAME,
        "auth_password": FUNCTIONAL_PASSWORD,
        "auth_secondary": FUNCTIONAL_PASSWORD,
        "auth_private_key_passphrase": FUNCTIONAL_PASSPHRASE,
        "auth_strict_key": False,
        "host": "172.18.0.11",
        "base_config": f"{TEST_DATA_PATH}/base_configs/cisco_iosxe",
    },
    "mock_cisco_iosxe": {
        "auth_username": MOCK_USERNAME,
        "auth_password": MOCK_PASSWORD,
        "auth_secondary": MOCK_PASSWORD,
        "auth_private_key_passphrase": MOCK_PASSPHRASE,
        "auth_strict_key": False,
        "host": "localhost",
        "port": 2211,
    },
    "cisco_nxos": {
        "auth_username": FUNCTIONAL_USERNAME,
        "auth_password": FUNCTIONAL_PASSWORD,
        "auth_secondary": FUNCTIONAL_PASSWORD,
        "auth_strict_key": False,
        "host": "172.18.0.12",
        "base_config": f"{TEST_DATA_PATH}/base_configs/cisco_nxos",
    },
    "mock_cisco_nxos": {
        "auth_username": MOCK_USERNAME,
        "auth_password": MOCK_PASSWORD,
        "auth_secondary": MOCK_PASSWORD,
        "auth_private_key_passphrase": MOCK_PASSPHRASE,
        "auth_strict_key": False,
        "host": "localhost",
        "port": 2212,
    },
    "cisco_iosxr": {
        "auth_username": FUNCTIONAL_USERNAME,
        "auth_password": FUNCTIONAL_PASSWORD,
        "auth_secondary": FUNCTIONAL_PASSWORD,
        "auth_strict_key": False,
        "host": "172.18.0.13",
        "base_config": f"{TEST_DATA_PATH}/base_configs/cisco_iosxr",
    },
    "arista_eos": {
        "auth_username": FUNCTIONAL_USERNAME,
        "auth_password": FUNCTIONAL_PASSWORD,
        "auth_secondary": FUNCTIONAL_PASSWORD,
        "auth_strict_key": False,
        "host": "172.18.0.14",
        "comms_ansi": True,
        "base_config": f"{TEST_DATA_PATH}/base_configs/arista_eos",
    },
    "juniper_junos": {
        "auth_username": FUNCTIONAL_USERNAME,
        "auth_password": FUNCTIONAL_PASSWORD,
        "auth_secondary": FUNCTIONAL_PASSWORD,
        "auth_strict_key": False,
        "host": "172.18.0.15",
        "base_config": f"{TEST_DATA_PATH}/base_configs/juniper_junos",
    },
    "mock_juniper_junos": {
        "auth_username": MOCK_USERNAME,
        "auth_password": MOCK_PASSWORD,
        "auth_secondary": MOCK_PASSWORD,
        "auth_private_key_passphrase": MOCK_PASSPHRASE,
        "auth_strict_key": False,
        "host": "localhost",
        "port": 2215,
    },
    "linux": {
        "auth_username": "root",
        "auth_password": "docker",
        "auth_strict_key": False,
        "host": "172.18.0.20",
        "comms_ansi": True,
        "comms_prompt_pattern": r"^linux:~#\s*$",
    },
}



import pytest

from scrapli.driver import GenericDriver

from .test_data.devices import DEVICES


@pytest.fixture(
    scope="class",
    params=["cisco_iosxe", "cisco_nxos", "cisco_iosxr", "arista_eos", "juniper_junos"],
)
def device_type(request):
    yield request.param


@pytest.fixture(scope="class", params=["system", "ssh2", "paramiko", "telnet"])
def transport(request):
    yield request.param


@pytest.fixture(scope="function", params=["asyncssh"])
def async_transport(request):
    yield request.param


@pytest.fixture(scope="class")
def nix_conn(transport):
    if transport == "telnet":
        pytest.skip("skipping telnet for linux hosts")

    device = DEVICES["linux"].copy()
    driver = device.pop("driver")
    device.pop("async_driver")

    conn = driver(
        **device,
        transport=transport,
    )
    conn.open()
    return conn


@pytest.fixture(scope="class")
def nix_conn_generic(transport):
    if transport == "telnet":
        pytest.skip("skipping telnet for linux hosts")

    device = DEVICES["linux"].copy()
    device.pop("driver")
    device.pop("async_driver")

    conn = GenericDriver(
        **device,
        transport=transport,
    )
    conn.open()
    return conn


@pytest.fixture(scope="class")
def conn(device_type, transport):
    if device_type == "arista_eos" and transport == "ssh2":
        pytest.skip(
            "SSH2 (on pypi) doesn't support keyboard interactive auth, skipping ssh2 for arista_eos testing"
        )

    device = DEVICES[device_type].copy()
    driver = device.pop("driver")
    device.pop("base_config")
    device.pop("async_driver")

    timeout_transport = 5
    timeout_ops = 5
    if device_type == "juniper_junos" or device_type == "cisco_iosxr":
        # commits on vsrx take one whole eternity... and iosxr container is just flakey
        timeout_transport = 30
        timeout_ops = 30

    port = 22
    if transport == "telnet":
        port = 23

    conn = driver(
        **device,
        port=port,
        transport=transport,
        timeout_socket=5,
        timeout_transport=timeout_transport,
        timeout_ops=timeout_ops,
    )
    conn.open()
    return conn


# scoping to function is probably dumb but dont have to screw around with which event loop is what this way
@pytest.fixture(scope="function")
async def async_conn(device_type, async_transport):
    device = DEVICES[device_type].copy()
    driver = device.pop("async_driver")
    device.pop("base_config")
    device.pop("driver")

    timeout_transport = 5
    timeout_ops = 5
    if device_type == "juniper_junos" or device_type == "cisco_iosxr":
        # commits on vsrx take one whole eternity... and iosxr container is just flakey
        timeout_transport = 30
        timeout_ops = 30

    async_conn = driver(
        **device,
        port=22,
        transport=async_transport,
        timeout_socket=5,
        timeout_transport=timeout_transport,
        timeout_ops=timeout_ops,
    )
    await async_conn.open()
    # yield then ensure we close since we are not persisting connections between tests for now
    yield async_conn
    if async_conn.isalive():
        await async_conn.close()


@pytest.fixture(scope="class")
def iosxe_conn(transport):
    device = DEVICES["cisco_iosxe"].copy()
    driver = device.pop("driver")
    device.pop("base_config")
    device.pop("async_driver")

    port = 22
    if transport == "telnet":
        port = 23

    iosxe_conn = driver(
        **device,
        port=port,
        transport=transport,
        timeout_socket=5,
    )
    iosxe_conn.open()
    return iosxe_conn


@pytest.fixture(scope="class")
def iosxr_conn(transport):
    device = DEVICES["cisco_iosxr"].copy()
    driver = device.pop("driver")
    device.pop("base_config")
    device.pop("async_driver")

    timeout_transport = 30
    timeout_ops = 30

    port = 22
    if transport == "telnet":
        port = 23

    conn = driver(
        **device,
        port=port,
        transport=transport,
        timeout_socket=5,
        timeout_transport=timeout_transport,
        timeout_ops=timeout_ops,
    )
    conn.open()
    return conn


@pytest.fixture(scope="class")
def junos_conn(transport):
    device = DEVICES["juniper_junos"].copy()
    driver = device.pop("driver")
    device.pop("base_config")
    device.pop("async_driver")

    timeout_transport = 30
    timeout_ops = 30

    port = 22
    if transport == "telnet":
        port = 23

    conn = driver(
        **device,
        port=port,
        transport=transport,
        timeout_socket=5,
        timeout_transport=timeout_transport,
        timeout_ops=timeout_ops,
    )
    conn.open()
    return conn


@pytest.fixture(scope="class")
def eos_conn(transport):
    if transport == "ssh2":
        pytest.skip(
            "SSH2 (on pypi) doesn't support keyboard interactive auth, skipping ssh2 for arista_eos testing"
        )
    device = DEVICES["arista_eos"].copy()
    driver = device.pop("driver")
    device.pop("base_config")
    device.pop("async_driver")

    port = 22
    if transport == "telnet":
        port = 23

    conn = driver(
        **device,
        port=port,
        transport=transport,
        timeout_socket=5,
    )
    conn.open()
    return conn


@pytest.fixture(scope="class")
def nxos_conn(transport):
    device = DEVICES["cisco_nxos"].copy()
    driver = device.pop("driver")
    device.pop("base_config")
    device.pop("async_driver")

    port = 22
    if transport == "telnet":
        port = 23

    conn = driver(
        **device,
        port=port,
        transport=transport,
        timeout_socket=5,
    )
    conn.open()
    return conn



import logging
import os
import sys
from pathlib import Path

import pytest

import scrapli
from scrapli import Scrape
from scrapli.exceptions import UnsupportedPlatform

TEST_DATA_DIR = f"{Path(scrapli.__file__).parents[1]}/tests/test_data"


def test__str():
    conn = Scrape(host="myhost")
    assert str(conn) == "Scrape Object for host myhost"


def test__repr():
    conn = Scrape(host="myhost")
    assert (
        repr(conn)
        == "Scrape(host='myhost', port=22, auth_username='', auth_password='', auth_private_key='', "
        "auth_private_key_passphrase='', auth_strict_key=True, auth_bypass=False, timeout_socket=5, "
        "timeout_transport=10, timeout_ops=30.0, timeout_exit=True, keepalive=False, keepalive_interval=30, "
        "keepalive_type='network', keepalive_pattern='\\x05', comms_prompt_pattern='^[a-z0-9.\\\\-@()/:]{1,"
        "48}[#>$]\\\\s*$', comms_return_char='\\n', comms_ansi=False, ssh_config_file='', ssh_known_hosts_file='', "
        "on_open=None, on_close=None, transport='system', transport_options=None)"
    )


@pytest.mark.parametrize(
    "attr_setup",
    [
        ("host", "", ValueError, "`host` should be a hostname/ip address, got nothing!"),
        ("port", "notanint", TypeError, "`port` should be int, got <class 'str'>"),
        (
            "auth_strict_key",
            "notabool",
            TypeError,
            "`auth_strict_key` should be bool, got <class 'str'>",
        ),
        (
            "auth_private_key",
            "notafile",
            ValueError,
            "Provided public key `notafile` is not a file",
        ),
        (
            "auth_bypass",
            "notabool",
            TypeError,
            "`auth_bypass` should be bool, got <class 'str'>",
        ),
        ("timeout_exit", "notabool", TypeError, "`timeout_exit` should be bool, got <class 'str'>"),
        ("keepalive", "notabool", TypeError, "`keepalive` should be bool, got <class 'str'>"),
        (
            "keepalive_type",
            "notvalid",
            ValueError,
            "`notvalid` is an invalid keepalive_type; must be 'network' or 'standard'",
        ),
        (
            "comms_return_char",
            True,
            TypeError,
            "`comms_return_char` should be str, got <class 'bool'>",
        ),
        ("comms_ansi", "notabool", TypeError, "`comms_ansi` should be bool, got <class 'str'>"),
        ("on_open", "notacallable", TypeError, "`on_open` must be a callable, got <class 'str'>"),
        ("on_close", "notacallable", TypeError, "`on_close` must be a callable, got <class 'str'>"),
        (
            "ssh_config_file",
            None,
            TypeError,
            "`ssh_config_file` must be str or bool, got <class 'NoneType'>",
        ),
        (
            "ssh_known_hosts_file",
            None,
            TypeError,
            "`ssh_known_hosts_file` must be str or bool, got <class 'NoneType'>",
        ),
    ],
    ids=[
        "host",
        "port",
        "auth_strict_key",
        "auth_private_key",
        "auth_bypass",
        "timeout_exit",
        "keepalive",
        "keepalive_type",
        "comms_return_char",
        "comms_ansi",
        "on_open",
        "on_close",
        "ssh_config_file",
        "ssh_known_hosts_file",
    ],
)
def test_exceptions_raised(attr_setup):
    attr_name = attr_setup[0]
    attr_value = attr_setup[1]
    attr_exc = attr_setup[2]
    attr_msg = attr_setup[3]
    args = {attr_name: attr_value}
    if attr_name != "host":
        args["host"] = "myhost"
    with pytest.raises(attr_exc) as exc:
        Scrape(**args)
    assert str(exc.value) == attr_msg


def test_unsupported_platform():
    original_platform = sys.platform
    sys.platform = "win32"
    with pytest.raises(UnsupportedPlatform) as exc:
        Scrape(host="localhost")
    assert (
        str(exc.value)
        == "`system` transport is not supported on Windows, please use a different transport"
    )
    sys.platform = original_platform


@pytest.mark.parametrize(
    "attr_setup",
    [
        ("host", "myhost", "myhost"),
        ("host", "myhost ", "myhost"),
        ("port", 123, 123),
        ("auth_username", "tacocat", "tacocat"),
        ("auth_username", "tacocat ", "tacocat"),
        ("auth_password", "tacocat", "tacocat"),
        ("auth_password", "tacocat ", "tacocat"),
        (
            "auth_private_key",
            f"{TEST_DATA_DIR}/files/_ssh_config",
            f"{TEST_DATA_DIR}/files/_ssh_config",
        ),
        ("auth_private_key_passphrase", "tacocat", "tacocat"),
        ("auth_private_key_passphrase", "tacocat ", "tacocat"),
        ("auth_strict_key", False, False),
        ("timeout_socket", 100, 100),
        ("timeout_transport", 100, 100),
        ("timeout_ops", 100, 100),
        ("timeout_exit", False, False),
        ("keepalive", True, True),
        ("keepalive_interval", 100, 100),
        ("keepalive_type", "standard", "standard"),
        ("keepalive_pattern", "tacocat", "tacocat"),
        ("comms_prompt_pattern", "tacocat", "tacocat"),
        ("comms_return_char", "tacocat", "tacocat"),
        ("comms_ansi", True, True),
        ("on_open", print, print),
        ("on_close", print, print),
        ("transport", "telnet", "telnet"),
    ],
    ids=[
        "host",
        "host_strip",
        "port",
        "auth_username",
        "auth_username_strip",
        "auth_password",
        "auth_password_strip",
        "auth_private_key",
        "auth_private_key_passphrase",
        "auth_private_key_passphrase_strip",
        "auth_strict_key",
        "timeout_socket",
        "timeout_transport",
        "timeout_ops",
        "timeout_exit",
        "keepalive",
        "keepalive_interval",
        "keepalive_type",
        "keepalive_pattern",
        "comms_prompt_pattern",
        "comms_return_char",
        "comms_ansi",
        "on_open",
        "on_close",
        "transport",
    ],
)
def test_attr_assignment(attr_setup):
    attr_name = attr_setup[0]
    attr_value = attr_setup[1]
    attr_expected = attr_setup[2]
    args = {attr_name: attr_value}
    if attr_name != "host":
        args["host"] = "myhost"
    conn = Scrape(**args)
    if attr_name == "transport":
        conn.transport_class == attr_expected
    else:
        assert conn._initialization_args.get(attr_name) == attr_expected


def test_log_non_core_transport(caplog):
    with caplog.at_level(logging.INFO):
        Scrape(host="myhost", transport="paramiko")
    assert "Non-core transport `paramiko` selected" in caplog.text


def test_transport_options():
    conn = Scrape(host="localhost", transport_options={"someoption": "somethingneat"})
    assert conn.transport_args["transport_options"] == {"someoption": "somethingneat"}


def test_valid_private_key_file():
    auth_private_key = f"{TEST_DATA_DIR}/files/_ssh_private_key"
    conn = Scrape(host="myhost", auth_private_key=auth_private_key)
    assert (
        conn._initialization_args["auth_private_key"] == f"{TEST_DATA_DIR}/files/_ssh_private_key"
    )


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="not currently testing ssh file resolution on windows"
)
@pytest.mark.parametrize(
    "ssh_file",
    [
        ("ssh_config_file", True, "/etc/ssh/ssh_config", f"{TEST_DATA_DIR}/files/_ssh_config"),
        (
            "ssh_config_file",
            True,
            f"{os.path.expanduser('~')}/.ssh/config",
            f"{TEST_DATA_DIR}/files/_ssh_config",
        ),
        (
            "ssh_config_file",
            f"{TEST_DATA_DIR}/files/_ssh_known_hosts",
            f"{TEST_DATA_DIR}/files/_ssh_known_hosts",
            f"{TEST_DATA_DIR}/files/_ssh_config",
        ),
        (
            "ssh_config_file",
            "",
            "",
            "",
        ),
        (
            "ssh_known_hosts_file",
            True,
            "/etc/ssh/ssh_known_hosts",
            f"{TEST_DATA_DIR}/files/_ssh_known_hosts",
        ),
        (
            "ssh_known_hosts_file",
            True,
            f"{os.path.expanduser('~')}/.ssh/known_hosts",
            f"{TEST_DATA_DIR}/files/_ssh_known_hosts",
        ),
        (
            "ssh_known_hosts_file",
            f"{TEST_DATA_DIR}/files/_ssh_known_hosts",
            f"{TEST_DATA_DIR}/files/_ssh_known_hosts",
            f"{TEST_DATA_DIR}/files/_ssh_known_hosts",
        ),
        (
            "ssh_known_hosts_file",
            "",
            "",
            "",
        ),
    ],
    ids=[
        "config_file_etc",
        "config_file_user",
        "config_file_manual",
        "config_file_not_found",
        "known_hosts_file_etc",
        "known_hosts_file_user",
        "known_hosts_file_manual",
        "known_hosts_file_not_found",
    ],
)
def test_ssh_files(fs, ssh_file):
    attr_name = ssh_file[0]
    attr_value = ssh_file[1]
    attr_expected = ssh_file[2]
    attr_src = ssh_file[3]
    args = {attr_name: attr_value}

    if attr_src and attr_expected:
        fs.add_real_file(source_path=attr_src, target_path=attr_expected)

    conn = Scrape(host="myhost", **args)
    assert conn._initialization_args[attr_name] == attr_expected


def test_isalive_no_transport():
    # test to ensure we handle attribute error to show that scrape is not alive if transport does
    # not exist yet
    conn = Scrape(host="myhost")
    assert conn.isalive() is False



import pytest

from scrapli.driver import AsyncNetworkDriver, NetworkDriver
from scrapli.driver.base_network_driver import PrivilegeLevel
from scrapli.driver.core import (
    AsyncEOSDriver,
    AsyncIOSXEDriver,
    AsyncIOSXRDriver,
    AsyncJunosDriver,
    AsyncNXOSDriver,
    EOSDriver,
    IOSXEDriver,
    IOSXRDriver,
    JunosDriver,
    NXOSDriver,
)
from scrapli.exceptions import ScrapliException
from scrapli.factory import AsyncScrapli, Scrapli

TEST_COMMUNITY_PRIV_LEVELS = {
    "exec": (
        PrivilegeLevel(
            pattern=r"^[a-z0-9.\-_@()/:]{1,63}>$",
            name="exec",
            previous_priv="",
            deescalate="",
            escalate="",
            escalate_auth=False,
            escalate_prompt="",
        )
    ),
    "privilege_exec": (
        PrivilegeLevel(
            pattern=r"^[a-z0-9.\-_@/:]{1,63}#$",
            name="privilege_exec",
            previous_priv="exec",
            deescalate="disable",
            escalate="enable",
            escalate_auth=True,
            escalate_prompt=r"^[pP]assword:\s?$",
        )
    ),
    "configuration": (
        PrivilegeLevel(
            pattern=r"^[a-z0-9.\-_@/:]{1,63}\(conf[a-z0-9.\-@/:\+]{0,32}\)#$",
            name="configuration",
            previous_priv="privilege_exec",
            deescalate="end",
            escalate="configure terminal",
            escalate_auth=False,
            escalate_prompt="",
        )
    ),
}


@pytest.mark.parametrize(
    "platform",
    [
        ("cisco_iosxe", IOSXEDriver),
        ("cisco_iosxr", IOSXRDriver),
        ("cisco_nxos", NXOSDriver),
        ("arista_eos", EOSDriver),
        ("juniper_junos", JunosDriver),
    ],
    ids=["cisco_iosxe", "cisco_iosxr", "cisco_nxos", "arista_eos", "juniper_junos"],
)
def test_sync_factory(platform):
    driver = Scrapli(platform=platform[0], host="localhost")
    assert isinstance(driver, platform[1])


@pytest.mark.parametrize(
    "platform",
    [
        ("cisco_iosxe", AsyncIOSXEDriver),
        ("cisco_iosxr", AsyncIOSXRDriver),
        ("cisco_nxos", AsyncNXOSDriver),
        ("arista_eos", AsyncEOSDriver),
        ("juniper_junos", AsyncJunosDriver),
    ],
    ids=["cisco_iosxe", "cisco_iosxr", "cisco_nxos", "arista_eos", "juniper_junos"],
)
def test_async_factory(platform):
    driver = AsyncScrapli(platform=platform[0], host="localhost", transport="asyncssh")
    assert isinstance(driver, platform[1])


def test_sync_factory_async_transport_exception():
    with pytest.raises(ScrapliException) as exc:
        Scrapli(platform="cisco_iosxe", transport="asyncssh")
    assert str(exc.value) == "Use `AsyncScrapli` if using an async transport!"


def test_async_factory_sync_transport_exception():
    with pytest.raises(ScrapliException) as exc:
        AsyncScrapli(platform="cisco_iosxe", transport="system")
    assert str(exc.value) == "Use `Scrapli` if using a synchronous transport!"


@pytest.mark.parametrize(
    "factory_setup",
    [(Scrapli, "system"), (AsyncScrapli, "asyncssh")],
    ids=["sync_factory", "async_factory"],
)
def test_factory_platform_bad_type(factory_setup):
    Factory = factory_setup[0]
    transport = factory_setup[1]
    with pytest.raises(ScrapliException) as exc:
        Factory(platform=True, transport=transport)
    assert str(exc.value) == "Argument `platform` must be `str` got `<class 'bool'>`"


@pytest.mark.parametrize(
    "factory_setup",
    [(Scrapli, "system", NetworkDriver), (AsyncScrapli, "asyncssh", AsyncNetworkDriver)],
    ids=["sync_factory", "async_factory"],
)
def test_factory_community_platform_defaults(factory_setup):
    Factory = factory_setup[0]
    transport = factory_setup[1]
    expected_driver = factory_setup[2]
    driver = Factory(platform="scrapli_networkdriver", host="localhost", transport=transport)
    assert isinstance(driver, expected_driver)
    assert driver._transport == transport
    assert driver.failed_when_contains == [
        "% Ambiguous command",
        "% Incomplete command",
        "% Invalid input detected",
        "% Unknown command",
    ]
    assert driver.textfsm_platform == "cisco_iosxe"
    assert driver.genie_platform == "iosxe"
    assert driver.default_desired_privilege_level == "privilege_exec"
    assert callable(driver.on_open)
    assert callable(driver.on_close)
    for actual_priv_level, expected_priv_level in zip(
        driver.privilege_levels.values(), TEST_COMMUNITY_PRIV_LEVELS.values()
    ):
        actual_priv_level.name == expected_priv_level.name
        actual_priv_level.pattern == expected_priv_level.pattern


@pytest.mark.parametrize(
    "factory_setup",
    [(Scrapli, "system", NetworkDriver), (AsyncScrapli, "asyncssh", AsyncNetworkDriver)],
    ids=["sync_factory", "async_factory"],
)
def test_sync_factory_community_platform_variant(factory_setup):
    Factory = factory_setup[0]
    transport = factory_setup[1]
    expected_driver = factory_setup[2]
    driver = Factory(
        platform="scrapli_networkdriver",
        host="localhost",
        variant="test_variant1",
        transport=transport,
    )
    assert isinstance(driver, expected_driver)
    assert driver._transport == transport
    assert driver.failed_when_contains == [
        "% Ambiguous command",
        "% Incomplete command",
        "% Invalid input detected",
        "% Unknown command",
    ]
    assert driver.textfsm_platform == "cisco_iosxe"
    assert driver.genie_platform == "iosxe"
    assert driver.default_desired_privilege_level == "configuration"
    assert callable(driver.on_open)
    assert callable(driver.on_close)
    for actual_priv_level, expected_priv_level in zip(
        driver.privilege_levels.values(), TEST_COMMUNITY_PRIV_LEVELS.values()
    ):
        actual_priv_level.name == expected_priv_level.name
        actual_priv_level.pattern == expected_priv_level.pattern



import logging
import time
from pathlib import Path

from device_info import junos_device

from scrapli.driver.core import JunosDriver

logging.basicConfig(
    filename=f"{Path(__file__).resolve().parents[0]}/junos_driver.log", level=logging.DEBUG
)
logger = logging.getLogger("scrapli")

conn = JunosDriver(**junos_device)
conn.open()

print("***** Get Prompt:")
prompt = conn.get_prompt()
print(prompt)

print("***** Show configuration | match 10.0.0.15:")
result = conn.send_command("show configuration | match 10.0.0.15")
print(result, result.result)

print("***** Show configuration:")
result = conn.send_command("show configuration")
print(result, result.result)

if junos_device["keepalive"]:
    print("***** Waiting for keepalive....")
    time.sleep(5)

conn.close()



"""scrapli.driver.core.cisco_iosxe.base_driver"""
from scrapli.driver.base_network_driver import PrivilegeLevel

PRIVS = {
    "exec": (
        PrivilegeLevel(
            pattern=r"^[a-z0-9.\-_@()/:]{1,63}>$",
            name="exec",
            previous_priv="",
            deescalate="",
            escalate="",
            escalate_auth=False,
            escalate_prompt="",
        )
    ),
    "privilege_exec": (
        PrivilegeLevel(
            pattern=r"^[a-z0-9.\-_@/:]{1,63}#$",
            name="privilege_exec",
            previous_priv="exec",
            deescalate="disable",
            escalate="enable",
            escalate_auth=True,
            escalate_prompt=r"^(?:enable\s){0,1}password:\s?$",
        )
    ),
    "configuration": (
        PrivilegeLevel(
            pattern=r"^[a-z0-9.\-_@/:]{1,63}\(conf[a-z0-9.\-@/:\+]{0,32}\)#$",
            name="configuration",
            previous_priv="privilege_exec",
            deescalate="end",
            escalate="configure terminal",
            escalate_auth=False,
            escalate_prompt="",
        )
    ),
}



from pathlib import Path

import scrapli
from scrapli import AsyncScrape, Scrape
from scrapli.driver.core import (
    AsyncEOSDriver,
    AsyncIOSXEDriver,
    AsyncIOSXRDriver,
    AsyncJunosDriver,
    AsyncNXOSDriver,
    EOSDriver,
    IOSXEDriver,
    IOSXRDriver,
    JunosDriver,
    NXOSDriver,
)

TEST_DATA_PATH = f"{Path(scrapli.__file__).parents[1]}/tests/functional/test_data"

USERNAME = "vrnetlab"
PASSWORD = "VR-netlab9"
PRIVATE_KEY = f"{TEST_DATA_PATH}/vrnetlab_key"
INVALID_PRIVATE_KEY = f"{TEST_DATA_PATH}/__init__.py"

DEVICES = {
    "cisco_iosxe": {
        "driver": IOSXEDriver,
        "async_driver": AsyncIOSXEDriver,
        "auth_username": USERNAME,
        "auth_password": PASSWORD,
        "auth_secondary": PASSWORD,
        "auth_strict_key": False,
        "host": "172.18.0.11",
        "base_config": f"{TEST_DATA_PATH}/base_configs/cisco_iosxe",
    },
    "cisco_nxos": {
        "driver": NXOSDriver,
        "async_driver": AsyncNXOSDriver,
        "auth_username": USERNAME,
        "auth_password": PASSWORD,
        "auth_secondary": PASSWORD,
        "auth_strict_key": False,
        "host": "172.18.0.12",
        "base_config": f"{TEST_DATA_PATH}/base_configs/cisco_nxos",
    },
    "cisco_iosxr": {
        "driver": IOSXRDriver,
        "async_driver": AsyncIOSXRDriver,
        "auth_username": USERNAME,
        "auth_password": PASSWORD,
        "auth_secondary": PASSWORD,
        "auth_strict_key": False,
        "host": "172.18.0.13",
        "base_config": f"{TEST_DATA_PATH}/base_configs/cisco_iosxr",
    },
    "arista_eos": {
        "driver": EOSDriver,
        "async_driver": AsyncEOSDriver,
        "auth_username": USERNAME,
        "auth_password": PASSWORD,
        "auth_secondary": PASSWORD,
        "auth_strict_key": False,
        "host": "172.18.0.14",
        "comms_ansi": True,
        "base_config": f"{TEST_DATA_PATH}/base_configs/arista_eos",
    },
    "juniper_junos": {
        "driver": JunosDriver,
        "async_driver": AsyncJunosDriver,
        "auth_username": USERNAME,
        "auth_password": PASSWORD,
        "auth_secondary": PASSWORD,
        "auth_strict_key": False,
        "host": "172.18.0.15",
        "base_config": f"{TEST_DATA_PATH}/base_configs/juniper_junos",
    },
    "linux": {
        "driver": Scrape,
        "async_driver": AsyncScrape,
        "auth_username": "root",
        "auth_password": "docker",
        "auth_strict_key": False,
        "host": "172.18.0.20",
        "comms_ansi": True,
        "comms_prompt_pattern": r"^linux:~#\s*$",
    },
}



"""scrapli.transport"""
from scrapli.transport.async_transport import AsyncTransport
from scrapli.transport.systemssh import SYSTEM_SSH_TRANSPORT_ARGS, SystemSSHTransport
from scrapli.transport.telnet import TELNET_TRANSPORT_ARGS, TelnetTransport
from scrapli.transport.transport import Transport

__all__ = (
    "AsyncTransport",
    "Transport",
    "SystemSSHTransport",
    "SYSTEM_SSH_TRANSPORT_ARGS",
    "TELNET_TRANSPORT_ARGS",
    "TelnetTransport",
)



"""scrapli.transport.systemssh"""
import re
from select import select
from typing import Any, Dict, Optional

from scrapli.decorators import OperationTimeout, requires_open_session
from scrapli.exceptions import ScrapliAuthenticationFailed
from scrapli.helper import get_prompt_pattern, strip_ansi
from scrapli.ssh_config import SSHConfig
from scrapli.transport.ptyprocess import PtyProcess
from scrapli.transport.transport import Transport

SYSTEM_SSH_TRANSPORT_ARGS = (
    "auth_username",
    "auth_private_key",
    "auth_private_key_passphrase",
    "auth_password",
    "auth_strict_key",
    "auth_bypass",
    "ssh_config_file",
    "ssh_known_hosts_file",
    "comms_prompt_pattern",
    "comms_return_char",
    "comms_ansi",
    "timeout_ops",
)


class SystemSSHTransport(Transport):
    def __init__(
        self,
        host: str = "",
        port: int = 22,
        auth_username: str = "",
        auth_private_key: str = "",
        auth_private_key_passphrase: str = "",
        auth_password: str = "",
        auth_strict_key: bool = True,
        auth_bypass: bool = False,
        timeout_socket: int = 5,
        timeout_transport: int = 5,
        timeout_ops: float = 10,
        timeout_exit: bool = True,
        keepalive: bool = False,
        keepalive_interval: int = 30,
        keepalive_type: str = "",
        keepalive_pattern: str = "\005",
        comms_prompt_pattern: str = r"^[a-z0-9.\-@()/:]{1,32}[#>$]$",
        comms_return_char: str = "\n",
        comms_ansi: bool = False,
        ssh_config_file: str = "",
        ssh_known_hosts_file: str = "",
        transport_options: Optional[Dict[str, Any]] = None,
    ) -> None:
        r"""
        SystemSSHTransport Object

        Inherit from Transport ABC
        SSH2Transport <- Transport (ABC)

        If using this driver, and passing a ssh_config_file (or setting this argument to `True`),
        all settings in the ssh config file will be superseded by any arguments passed here!

        Note that comms_prompt_pattern, comms_return_char and comms_ansi are only passed here to
        handle "in channel" authentication required by SystemSSH -- these are assigned to private
        attributes in this class and ignored after authentication. If you wish to modify these
        values on a "live" scrapli connection, modify them in the Channel object, i.e.
        `conn.channel.comms_prompt_pattern`. Additionally timeout_ops is passed and assigned to
        _timeout_ops to use the same timeout_ops that is used in Channel to decorate the
        authentication methods here.

        Args:
            host: host ip/name to connect to
            port: port to connect to
            auth_username: username for authentication
            auth_private_key: path to private key for authentication
            auth_private_key_passphrase: passphrase for decrypting ssh key if necessary
            auth_password: password for authentication
            auth_strict_key: True/False to enforce strict key checking (default is True)
            auth_bypass: bypass ssh key or password auth for devices without authentication, or that
                have auth prompts after ssh session establishment
            timeout_socket: timeout for ssh session to start -- this directly maps to ConnectTimeout
                ssh argument; see `man ssh_config`
            timeout_transport: timeout for transport in seconds. since system ssh is using popen/pty
                we can't really set a timeout directly, so this value governs the time timeout
                decorator for the transport read and write methods
            timeout_ops: timeout for telnet channel operations in seconds -- this is also the
                timeout for finding and responding to username and password prompts at initial
                login. This is assigned to a private attribute and is ignored after authentication
                is completed.
            timeout_exit: True/False close transport if timeout encountered. If False and keepalives
                are in use, keepalives will prevent program from exiting so you should be sure to
                catch Timeout exceptions and handle them appropriately
            keepalive: whether or not to try to keep session alive
            keepalive_interval: interval to use for session keepalives
            keepalive_type: network|standard -- 'network' sends actual characters over the
                transport channel. This is useful for network-y type devices that may not support
                'standard' keepalive mechanisms. 'standard' is not currently implemented for
                system ssh
            keepalive_pattern: pattern to send to keep network channel alive. Default is
                u'\005' which is equivalent to 'ctrl+e'. This pattern moves cursor to end of the
                line which should be an innocuous pattern. This will only be entered *if* a lock
                can be acquired. This is only applicable if using keepalives and if the keepalive
                type is 'network'
            comms_prompt_pattern: prompt pattern expected for device, same as the one provided to
                channel -- system ssh needs to know this to know how to decide if we are properly
                sending/receiving data -- i.e. we are not stuck at some password prompt or some
                other failure scenario. If using driver, this should be passed from driver (Scrape,
                or IOSXE, etc.) to this Transport class. This is assigned to a private attribute and
                is ignored after authentication is completed.
            comms_return_char: return character to use on the channel, same as the one provided to
                channel -- system ssh needs to know this to know what to send so that we can probe
                the channel to make sure we are authenticated and sending/receiving data. If using
                driver, this should be passed from driver (Scrape, or IOSXE, etc.) to this Transport
                class. This is assigned to a private attribute and is ignored after authentication
                is completed.
            comms_ansi: True/False strip comms_ansi characters from output; this value is assigned
                self._comms_ansi and is ignored after authentication. We only need it for transport
                on the off chance (maybe never?) that username/password prompts contain ansi
                characters, otherwise "comms_ansi" is really a channel attribute and is treated as
                such. This is assigned to a private attribute and is ignored after authentication
                is completed.
            ssh_config_file: string to path for ssh config file
            ssh_known_hosts_file: string to path for ssh known hosts file
            transport_options: SystemSSHTransport specific transport options (options that don't
                apply to any of the other transport classes) supplied in a dictionary where the key
                is the name of the option and the value is of course the value.
                - open_cmd: string or list of strings to extend the open_cmd with, for example:
                    `["-o", "KexAlgorithms=+diffie-hellman-group1-sha1"]` or:
                    `-oKexAlgorithms=+diffie-hellman-group1-sha1`
                    these commands will be appended to the open command that scrapli builds which
                    looks something like the following depending on the inputs provided:
                        ssh 172.31.254.1 -p 22 -o ConnectTimeout=5 -o ServerAliveInterval=10
                         -l scrapli -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null
                         -F /dev/null
                    You can pass any arguments that would be supported if you were ssh'ing on your
                    terminal "normally", passing some bad arguments can break things!

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        super().__init__(
            host,
            port,
            timeout_socket,
            timeout_transport,
            timeout_exit,
            keepalive,
            keepalive_interval,
            keepalive_type,
            keepalive_pattern,
        )

        self.auth_username: str = auth_username
        self.auth_private_key: str = auth_private_key
        self.auth_private_key_passphrase: str = auth_private_key_passphrase
        self.auth_password: str = auth_password
        self.auth_strict_key: bool = auth_strict_key
        self.auth_bypass: bool = auth_bypass

        self._timeout_ops: float = timeout_ops
        self._comms_prompt_pattern: str = comms_prompt_pattern
        self._comms_return_char: str = comms_return_char
        self._comms_ansi: bool = comms_ansi
        self._process_ssh_config(ssh_config_file)
        self.ssh_known_hosts_file: str = ssh_known_hosts_file

        self.session: PtyProcess
        self.lib_auth_exception = ScrapliAuthenticationFailed
        self._isauthenticated = False

        # ensure we set transport_options to a dict if its left as None
        self.transport_options = transport_options or {}

        self.open_cmd = ["ssh", self.host]
        self._build_open_cmd()

    def _process_ssh_config(self, ssh_config_file: str) -> None:
        """
        Method to parse ssh config file

        Ensure ssh_config_file is valid (if providing a string path to config file), or resolve
        config file if passed True.

        Args:
            ssh_config_file: string path to ssh config file; passed down from `Scrape`, or the
                `NetworkDriver` or subclasses of it, in most cases.

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        ssh = SSHConfig(ssh_config_file=ssh_config_file)
        self.ssh_config_file = ssh.ssh_config_file

    def _build_open_cmd(self) -> None:
        """
        Method to craft command to open ssh session

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.open_cmd.extend(["-p", str(self.port)])
        self.open_cmd.extend(["-o", f"ConnectTimeout={self.timeout_socket}"])
        self.open_cmd.extend(["-o", f"ServerAliveInterval={self.timeout_transport}"])
        if self.auth_private_key:
            self.open_cmd.extend(["-i", self.auth_private_key])
        if self.auth_username:
            self.open_cmd.extend(["-l", self.auth_username])
        if self.auth_strict_key is False:
            self.open_cmd.extend(["-o", "StrictHostKeyChecking=no"])
            self.open_cmd.extend(["-o", "UserKnownHostsFile=/dev/null"])
        else:
            self.open_cmd.extend(["-o", "StrictHostKeyChecking=yes"])
            if self.ssh_known_hosts_file:
                self.open_cmd.extend(["-o", f"UserKnownHostsFile={self.ssh_known_hosts_file}"])
        if self.ssh_config_file:
            self.open_cmd.extend(["-F", self.ssh_config_file])
        else:
            self.open_cmd.extend(["-F", "/dev/null"])

        user_args = self.transport_options.get("open_cmd", [])
        if isinstance(user_args, str):
            user_args = [user_args]
        self.open_cmd.extend(user_args)

    def open(self) -> None:
        """
        Parent method to open session, authenticate and acquire shell

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.logger.info(f"Attempting to authenticate to {self.host}")
        self._open_pty()
        self.logger.info(f"Successfully authenticated to {self.host}")

        if self.keepalive:
            self._session_keepalive()

    def _ssh_message_handler(self, output: bytes) -> None:  # noqa: C901
        """
        Parse EOF messages from _pty_authenticate and create log/stack exception message

        Args:
            output: bytes output from _pty_authenticate

        Returns:
            N/A  # noqa: DAR202

        Raises:
            ScrapliAuthenticationFailed: if any errors are read in the output

        """
        msg = ""
        if b"host key verification failed" in output.lower():
            msg = f"Host key verification failed for host {self.host}"
        elif b"operation timed out" in output.lower() or b"connection timed out" in output.lower():
            msg = f"Timed out connecting to host {self.host}"
        elif b"no route to host" in output.lower():
            msg = f"No route to host {self.host}"
        elif b"no matching key exchange" in output.lower():
            msg = f"No matching key exchange found for host {self.host}"
            key_exchange_pattern = re.compile(
                pattern=rb"their offer: ([a-z0-9\-,]*)", flags=re.M | re.I
            )
            offered_key_exchanges_match = re.search(pattern=key_exchange_pattern, string=output)
            if offered_key_exchanges_match:
                offered_key_exchanges = offered_key_exchanges_match.group(1).decode()
                msg = (
                    f"No matching key exchange found for host {self.host}, their offer: "
                    f"{offered_key_exchanges}"
                )
        elif b"no matching cipher" in output.lower():
            msg = f"No matching cipher found for host {self.host}"
            ciphers_pattern = re.compile(pattern=rb"their offer: ([a-z0-9\-,]*)", flags=re.M | re.I)
            offered_ciphers_match = re.search(pattern=ciphers_pattern, string=output)
            if offered_ciphers_match:
                offered_ciphers = offered_ciphers_match.group(1).decode()
                msg = (
                    f"No matching cipher found for host {self.host}, their offer: {offered_ciphers}"
                )
        elif b"bad configuration" in output.lower():
            msg = f"Bad SSH configuration option(s) for host {self.host}"
            configuration_pattern = re.compile(
                pattern=rb"bad configuration option: ([a-z0-9\+\=,]*)", flags=re.M | re.I
            )
            configuration_issue_match = re.search(pattern=configuration_pattern, string=output)
            if configuration_issue_match:
                configuration_issues = configuration_issue_match.group(1).decode()
                msg = (
                    f"Bad SSH configuration option(s) for host {self.host}, bad option(s): "
                    f"{configuration_issues}"
                )
        elif b"WARNING: UNPROTECTED PRIVATE KEY FILE!" in output:
            msg = (
                f"Permissions for private key `{self.auth_private_key}` are too open, "
                "authentication failed!"
            )
        elif b"could not resolve hostname" in output.lower():
            msg = f"Could not resolve address for host `{self.host}`"
        if msg:
            self.logger.critical(msg)
            raise ScrapliAuthenticationFailed(msg)

    def _open_pty(self) -> bool:
        """
        Private method to open session with PtyProcess

        Args:
            N/A

        Returns:
            bool: True/False session was opened and authenticated

        Raises:
            ScrapliAuthenticationFailed: if all authentication means fail

        """
        self.logger.info(f"Attempting to open session with the following command: {self.open_cmd}")
        self.session = PtyProcess.spawn(self.open_cmd)
        self.logger.debug(f"Session to host {self.host} spawned")

        if not self.auth_bypass:
            self._authenticate()
        else:
            self.logger.info("`auth_bypass` is True, bypassing authentication")
            # if we skip auth, we'll manually set _isauthenticated to True
            self._isauthenticated = True

        if not self._isauthenticated and not self._system_isauthenticated():
            msg = f"Authentication to host {self.host} failed"
            self.logger.critical(msg)
            raise ScrapliAuthenticationFailed(msg)

        self.logger.debug(f"Authenticated to host {self.host} with password")
        return True

    @OperationTimeout("_timeout_ops", "Timed out attempting to authenticate")
    def _authenticate(self) -> None:
        """
        Private method to check initial authentication when using pty_session

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            ScrapliAuthenticationFailed: if we get EOF and _ssh_message_handler does not raise an
                explicit exception/message

        """
        output = b""
        prompt_pattern = get_prompt_pattern(prompt="", class_prompt=self._comms_prompt_pattern)
        with self.session_lock:
            while True:
                try:
                    new_output = self.session.read()
                    output += new_output
                    self.logger.debug(f"Attempting to authenticate. Read: {repr(new_output)}")
                except EOFError as exc:
                    self._ssh_message_handler(output=output)
                    # if _ssh_message_handler didn't raise any exception, we can raise the standard
                    # "did you disable strict key message/exception"
                    msg = (
                        f"Failed to open connection to host {self.host}. Do you need to disable "
                        "`auth_strict_key`?"
                    )
                    self.logger.critical(msg)
                    raise ScrapliAuthenticationFailed(msg) from exc

                # even if we dont hit EOF, we may still have hit a message we need to process such
                # as insecure key permissions
                self._ssh_message_handler(output=output)

                if self._comms_ansi or b"\x1B" in output:
                    output = strip_ansi(output=output)
                if b"password" in output.lower():
                    self.logger.info("Found password prompt, sending password")
                    self.session.write(self.auth_password.encode())
                    self.session.write(self._comms_return_char.encode())
                    break
                if b"enter passphrase for key" in output.lower():
                    self.logger.info("Found key passphrase prompt, sending passphrase")
                    self.session.write(self.auth_private_key_passphrase.encode())
                    self.session.write(self._comms_return_char.encode())
                    break
                channel_match = re.search(pattern=prompt_pattern, string=output)
                if channel_match:
                    self.logger.info(
                        "Found channel prompt, assuming that key based authentication has succeeded"
                    )
                    # set _isauthenticated to true, we've already authenticated, don't re-check
                    self._isauthenticated = True
                    break

    @OperationTimeout("_timeout_ops", "Timed out determining if session is authenticated")
    def _system_isauthenticated(self) -> bool:
        """
        Check if session is authenticated

        This is very naive -- it only knows if the sub process is alive and has not received an EOF.
        Beyond that we lock the session and send the return character and re-read the channel.

        Args:
            N/A

        Returns:
            bool: True if authenticated, else False

        Raises:
            N/A

        """
        self.logger.debug("Attempting to determine if authentication was successful")
        with self.session_lock:
            if self.session.isalive() and not self.session.eof():
                prompt_pattern = get_prompt_pattern(
                    prompt="", class_prompt=self._comms_prompt_pattern
                )
                self.session.write(self._comms_return_char.encode())
                while True:
                    # almost all of the time we don't need a while loop here, but every once in a
                    # while fd won't be ready which causes a failure without an obvious root cause,
                    # loop/logging to hopefully help with that
                    fd_ready, _, _ = select([self.session.fd], [], [], 0)
                    if self.session.fd in fd_ready:
                        break
                    self.logger.debug("PTY fd not ready yet...")
                output = b""
                while True:
                    new_output = self.session.read()
                    output += new_output
                    self.logger.debug(
                        f"Attempting to validate authentication. Read: {repr(new_output)}"
                    )
                    # we do not need to deal w/ line replacement for the actual output, only for
                    # parsing if a prompt-like thing is at the end of the output
                    output = output.replace(b"\r", b"")
                    # always check to see if we should strip ansi here; if we don't handle this we
                    # may raise auth failures for the wrong reason which would be confusing for
                    # users
                    if self._comms_ansi or b"\x1B" in output:
                        output = strip_ansi(output=output)
                    channel_match = re.search(pattern=prompt_pattern, string=output)
                    if channel_match:
                        self._isauthenticated = True
                        break
                    if b"password:" in output.lower():
                        # if we see "password" we know auth failed (hopefully in all scenarios!)
                        self.logger.critical(
                            "Found `password:` in output, assuming password authentication failed"
                        )
                        break
                    if output:
                        self.logger.debug(
                            f"Cannot determine if authenticated, \n\tRead: {repr(output)}"
                        )

        if self._isauthenticated:
            return True

        return False

    def close(self) -> None:
        """
        Close session and socket

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        with self.session_lock:
            self.session.kill(0)
            self.session.terminated = True
            self.logger.debug(f"Channel to host {self.host} closed")

    def isalive(self) -> bool:
        """
        Check if session is alive and session is authenticated

        Args:
            N/A

        Returns:
            bool: True if session is alive and session authenticated, else False

        Raises:
            N/A

        """
        if self.session.isalive() and self._isauthenticated and not self.session.eof():
            return True
        return False

    @requires_open_session()
    @OperationTimeout("timeout_transport", "Timed out reading from transport")
    def read(self) -> bytes:
        """
        Read data from the channel

        Args:
            N/A

        Returns:
            bytes: bytes output as read from channel

        Raises:
            N/A

        """
        read_bytes = 65535
        return self.session.read(read_bytes)

    @requires_open_session()
    @OperationTimeout("timeout_transport", "Timed out writing to transport")
    def write(self, channel_input: str) -> None:
        """
        Write data to the channel

        Args:
            channel_input: string to send to channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.session.write(channel_input.encode())

    def set_timeout(self, timeout: Optional[int] = None) -> None:
        """
        Set session timeout

        Note that this modifies the objects `timeout_transport` value directly as this value is
        what controls the timeout decorator for read/write methods. This is slightly different
        behavior from ssh2/paramiko/telnet in that those transports modify the session value and
        leave the objects `timeout_transport` alone.

        Args:
            timeout: timeout in seconds

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if isinstance(timeout, int):
            set_timeout = timeout
        else:
            set_timeout = self.timeout_transport
        self.timeout_transport = set_timeout

    def _keepalive_standard(self) -> None:
        """
        Send "out of band" (protocol level) keepalives to devices.

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            NotImplementedError: 'standard' keepalive mechanism not yet implemented for system

        """
        raise NotImplementedError("'standard' keepalive mechanism not yet implemented for system.")



"""scrapli.driver.core.cisco_nxos.async_driver"""
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional

from scrapli.driver import AsyncNetworkDriver
from scrapli.driver.base_network_driver import PrivilegeLevel
from scrapli.driver.core.cisco_nxos.base_driver import PRIVS, NXOSDriverBase


async def nxos_on_open(conn: AsyncNetworkDriver) -> None:
    """
    AsyncNXOSDriver default on_open callable

    Args:
        conn: NetworkDriver object

    Returns:
        N/A  # noqa: DAR202

    Raises:
        N/A
    """
    await conn.send_command(command="terminal length 0")
    await conn.send_command(command="terminal width 511")


async def nxos_on_close(conn: AsyncNetworkDriver) -> None:
    """
    AsyncNXOSDriver default on_close callable

    Args:
        conn: NetworkDriver object

    Returns:
        N/A  # noqa: DAR202

    Raises:
        N/A
    """
    # write exit directly to the transport as channel would fail to find the prompt after sending
    # the exit command!
    await conn.acquire_priv(desired_priv=conn.default_desired_privilege_level)
    conn.transport.write(channel_input="exit")
    conn.transport.write(channel_input=conn.channel.comms_return_char)


class AsyncNXOSDriver(AsyncNetworkDriver, NXOSDriverBase):
    def __init__(
        self,
        privilege_levels: Optional[Dict[str, PrivilegeLevel]] = None,
        default_desired_privilege_level: str = "privilege_exec",
        auth_secondary: str = "",
        on_open: Optional[Callable[..., Any]] = None,
        on_close: Optional[Callable[..., Any]] = None,
        textfsm_platform: str = "cisco_nxos",
        genie_platform: str = "nxos",
        failed_when_contains: Optional[List[str]] = None,
        transport: str = "system",
        **kwargs: Dict[str, Any],
    ):
        """
        NXOSDriver Object

        Args:
            privilege_levels: optional user provided privilege levels, if left None will default to
                scrapli standard privilege levels
            default_desired_privilege_level: string of name of default desired priv, this is the
                priv level that is generally used to disable paging/set terminal width and things
                like that upon first login, and is also the priv level scrapli will try to acquire
                for normal "command" operations (`send_command`, `send_commands`)
            auth_secondary: password to use for secondary authentication (enable)
            on_open: callable that accepts the class instance as its only argument. this callable,
                if provided, is executed immediately after authentication is completed. Common use
                cases for this callable would be to disable paging or accept any kind of banner
                message that prompts a user upon connection
            on_close: callable that accepts the class instance as its only argument. this callable,
                if provided, is executed immediately prior to closing the underlying transport.
                Common use cases for this callable would be to save configurations prior to exiting,
                or to logout properly to free up vtys or similar.
            textfsm_platform: string name of textfsm parser platform
            genie_platform: string name of cisco genie parser platform
            failed_when_contains: List of strings that indicate a command/config has failed
            transport: system|telnet or a plugin -- type of transport to use for connection
                system uses system available ssh (/usr/bin/ssh)
                ssh2 uses ssh2-python *has been migrated to a plugin
                paramiko uses... paramiko *has been migrated to a plugin
                telnet uses telnetlib
                choice of driver depends on the features you need. in general system is easiest as
                it will just 'auto-magically' use your ssh config file ('~/.ssh/config' or
                '/etc/ssh/config_file'). ssh2 is very very fast as it is a thin wrapper around
                libssh2 however it is slightly feature limited. paramiko is slower than ssh2, but
                has more features built in (though scrapli does not expose/support them all).
                explicitly added here to allow for nicely checking if transport is telnet.
            **kwargs: keyword args to pass to inherited class(es)

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A
        """
        if privilege_levels is None:
            privilege_levels = deepcopy(PRIVS)

        if on_open is None:
            on_open = nxos_on_open
        if on_close is None:
            on_close = nxos_on_close

        _telnet = False
        if transport == "telnet":
            _telnet = True

        if failed_when_contains is None:
            failed_when_contains = [
                "% Ambiguous command",
                "% Incomplete command",
                "% Invalid input detected",
                "% Invalid command at",
            ]

        super().__init__(
            privilege_levels=privilege_levels,
            default_desired_privilege_level=default_desired_privilege_level,
            auth_secondary=auth_secondary,
            failed_when_contains=failed_when_contains,
            textfsm_platform=textfsm_platform,
            genie_platform=genie_platform,
            on_open=on_open,
            on_close=on_close,
            transport=transport,
            **kwargs,
        )

        if _telnet:
            self.transport.username_prompt = "login:"

    async def _abort_config(self) -> None:
        """
        Abort NXOS configuration session (if using a config session!)

        Args:
            N/A

        Returns:
            N/A:  # noqa: DAR202

        Raises:
            N/A

        """
        # nxos pattern for config sessions should *always* have `config-s`
        if "config\\-s" in self._current_priv_level.pattern:
            await self.channel.send_input(channel_input="abort")
            self._current_priv_level = self.privilege_levels["privilege_exec"]

    def register_configuration_session(self, session_name: str) -> None:
        """
        NXOS specific implementation of register_configuration_session

        Args:
            session_name: name of session to register

        Returns:
            N/A:  # noqa: DAR202

        Raises:
            N/A

        """
        self._create_configuration_session(session_name=session_name)
        self.update_privilege_levels(update_channel=True)



#!/usr/bin/env python
"""scrapli - ssh|telnet screen scraping client library"""
import setuptools

from scrapli import __version__

__author__ = "Carl Montanari"

with open("README.md", "r", encoding="utf-8") as f:
    README = f.read()

setuptools.setup(
    name="scrapli",
    version=__version__,
    author=__author__,
    author_email="carl.r.montanari@gmail.com",
    description="Screen scraping (ssh|telnet) client focused on network devices",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/carlmontanari/scrapli",
    packages=setuptools.find_packages(),
    install_requires=[],
    extras_require={
        "textfsm": ["textfsm>=1.1.0", "ntc-templates>=1.1.0"],
        "genie": ["genie>=20.2", "pyats>=20.2"],
        "paramiko": ["scrapli_paramiko>=2020.06.06"],
        "ssh2": ["scrapli_ssh2>=2020.06.06"],
        "asyncssh": ["scrapli_asyncssh>=2020.07.04"],
        "netconf": ["scrapli_netconf>=2020.07.26"],
        "full": [
            "textfsm>=1.1.0",
            "ntc-templates>=1.1.0",
            "genie>=20.2",
            "pyats>=20.2",
            "scrapli_paramiko>=2020.06.06",
            "scrapli_ssh2>=2020.06.06",
            "scrapli_asyncssh>=2020.07.04",
            "scrapli_netconf>=2020.07.26",
        ],
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
    ],
    python_requires=">=3.6",
)



def test_configuration_session(eos_conn):
    eos_conn.register_configuration_session(session_name="scrapli_test_session1")
    result = eos_conn.send_configs(
        configs=["interface ethernet 1", "show configuration sessions"],
        privilege_level="scrapli_test_session1",
    )
    eos_conn.close()
    # pop the config session out
    eos_conn.privilege_levels.pop("scrapli_test_session1")
    assert "* scrapli_test_session" in result[1].result


def test_configuration_session_abort(eos_conn):
    eos_conn.register_configuration_session(session_name="scrapli_test_session2")
    result = eos_conn.send_configs(
        configs=["tacocat", "show configuration sessions"],
        privilege_level="scrapli_test_session2",
        stop_on_failed=True,
    )
    current_prompt = eos_conn.get_prompt()
    eos_conn.close()
    # pop the config session out
    eos_conn.privilege_levels.pop("scrapli_test_session2")
    # assert config session was aborted at first sign of failed config
    assert len(result) == 1
    # assert that session aborted and we are back at priv exec
    assert current_prompt == "localhost#"



import logging
import time
from pathlib import Path

from device_info import iosxe_device

from scrapli import Scrape

logging.basicConfig(
    filename=f"{Path(__file__).resolve().parents[0]}/scrape_driver.log", level=logging.DEBUG
)
logger = logging.getLogger("scrapli")

conn = Scrape(**iosxe_device)
conn.open()

print("***** Get Prompt:")
print(conn.channel.get_prompt())

print("***** Show run | i hostname:")
result = conn.channel.send_input("show run | i hostname")
print(result)

print("***** Clear logging buffer:")
interact = [("clear logg", "Clear logging buffer [confirm]"), ("", "csr1000v#")]
result = conn.channel.send_inputs_interact(interact)
print(result)

print("***** Disable Paging:")
result = conn.channel.send_input("term length 0")
print(result)

print("***** Show run:")
result = conn.channel.send_input("show run")
print(result)

if iosxe_device["keepalive"]:
    print("***** Waiting for keepalive....")
    time.sleep(5)

conn.close()



"""scrapli.transport.base_transport"""
from abc import ABC, abstractmethod
from logging import getLogger
from threading import Lock
from typing import Optional


class TransportBase(ABC):
    def __init__(
        self,
        host: str = "",
        port: int = 22,
        timeout_socket: int = 5,
        timeout_transport: int = 5,
        timeout_exit: bool = True,
        keepalive: bool = False,
        keepalive_interval: int = 30,
        keepalive_type: str = "network",
        keepalive_pattern: str = "\005",
    ) -> None:
        r"""
        Transport Base Object

        Args:
            host: host ip/name to connect to
            port: port to connect to
            timeout_socket: timeout for establishing socket in seconds
            timeout_transport: timeout for ssh|telnet transport in seconds
            timeout_exit: True/False close transport if timeout encountered. If False and keepalives
                are in use, keepalives will prevent program from exiting so you should be sure to
                catch Timeout exceptions and handle them appropriately
            keepalive: whether or not to try to keep session alive
            keepalive_interval: interval to use for session keepalives
            keepalive_type: network|standard -- 'network' sends actual characters over the
                transport channel. This is useful for network-y type devices that may not support
                'standard' keepalive mechanisms. 'standard' attempts to use whatever 'standard'
                keepalive mechanisms are available in the selected transport mechanism. Check the
                transport documentation for details on what is supported and/or how it is
                implemented for any given transport driver
            keepalive_pattern: pattern to send to keep network channel alive. Default is
                u'\005' which is equivalent to 'ctrl+e'. This pattern moves cursor to end of the
                line which should be an innocuous pattern. This will only be entered *if* a lock
                can be acquired. This is only applicable if using keepalives and if the keepalive
                type is 'network'

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.logger = getLogger(f"scrapli.transport-{host}")

        self.host: str = host
        self.port: int = port
        self.timeout_socket: int = timeout_socket
        self.timeout_transport: int = timeout_transport
        self.timeout_exit: bool = timeout_exit
        self.keepalive: bool = keepalive
        self.keepalive_interval: int = keepalive_interval
        self.keepalive_type: str = keepalive_type
        self.keepalive_pattern: str = keepalive_pattern

        self.session_lock: Lock = Lock()

    def __bool__(self) -> bool:
        """
        Magic bool method for Socket

        Args:
            N/A

        Returns:
            bool: True/False if socket is alive or not

        Raises:
            N/A

        """
        return self.isalive()

    def __str__(self) -> str:
        """
        Magic str method for Transport

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        return f"Transport Object for host {self.host}"

    def __repr__(self) -> str:
        """
        Magic repr method for Transport

        Args:
            N/A

        Returns:
            str: repr for class object

        Raises:
            N/A

        """
        class_dict = self.__dict__.copy()
        if "auth_password" in class_dict.keys():
            class_dict["auth_password"] = "********"
        if "auth_private_key_passphrase" in class_dict.keys():
            class_dict["auth_private_key_passphrase"] = "********"
        class_dict["session_lock"] = self.session_lock.locked()
        class_dict["logger"] = self.logger.name
        return f"Transport {class_dict}"

    @abstractmethod
    def close(self) -> None:
        """
        Close session and socket

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    @abstractmethod
    def isalive(self) -> bool:
        """
        Check if socket is alive and session is authenticated

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    @abstractmethod
    def set_timeout(self, timeout: Optional[int] = None) -> None:
        """
        Set session timeout

        Args:
            timeout: timeout in seconds

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """



"""
Ptyprocess is under the ISC license, as code derived from Pexpect.
    http://opensource.org/licenses/ISC

Copyright (c) 2013-2014, Pexpect development team
Copyright (c) 2012, Noah Spurrier <noah@noah.org>

PERMISSION TO USE, COPY, MODIFY, AND/OR DISTRIBUTE THIS SOFTWARE FOR ANY PURPOSE
WITH OR WITHOUT FEE IS HEREBY GRANTED, PROVIDED THAT THE ABOVE COPYRIGHT NOTICE
AND THIS PERMISSION NOTICE APPEAR IN ALL COPIES. THE SOFTWARE IS PROVIDED
"AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Thank you to the folks who create and maintain pexpect, vendoring this here
has been ridiculously helpful! -CM
"""

import builtins
import errno
import io
import os
import signal
import struct
import sys
import time
from shutil import which
from typing import List, Optional, Type, TypeVar

from scrapli.exceptions import SSHNotFound


class PtyProcessError(Exception):
    """Generic error class for this package."""


PtyProcessType = TypeVar("PtyProcessType", bound="PtyProcess")


_EOF, _INTR = None, None


def _make_eof_intr() -> None:
    """
    Set constants _EOF and _INTR.

    This avoids doing potentially costly operations on module load.

    Args:
        N/A

    Returns:
        N/A  # noqa: DAR202

    Raises:
        N/A

    """
    import termios

    global _EOF, _INTR
    if (_EOF is not None) and (_INTR is not None):
        return

    # inherit EOF and INTR definitions from controlling process.
    try:
        from termios import VEOF, VINTR

        fd = None
        for name in "stdin", "stdout":
            stream = getattr(sys, "__%s__" % name, None)
            if stream is None or not hasattr(stream, "fileno"):
                continue
            try:
                fd = stream.fileno()
            except ValueError:
                continue
        if fd is None:
            # no fd, raise ValueError to fallback on CEOF, CINTR
            raise ValueError("No stream has a fileno")
        intr = ord(termios.tcgetattr(fd)[6][VINTR])
        eof = ord(termios.tcgetattr(fd)[6][VEOF])
    except (ImportError, OSError, IOError, ValueError, termios.error):
        # unless the controlling process is also not a terminal,
        # such as cron(1), or when stdin and stdout are both closed.
        # Fall-back to using CEOF and CINTR. There
        try:
            from termios import CEOF, CINTR

            (intr, eof) = (CINTR, CEOF)
        except ImportError:
            (intr, eof) = (3, 4)

    _INTR = bytes([intr])
    _EOF = bytes([eof])


def _setwinsize(fd: int, rows: int, cols: int) -> None:
    """
    Set window size.

    Some very old platforms have a bug that causes the value for termios.TIOCSWINSZ to be truncated.
    There was a hack here to work around this, but it caused problems with newer platforms so has
    been removed. For details see https://github.com/pexpect/pexpect/issues/39

    Args:
        fd: file descriptor
        rows: int number of rows for terminal
        cols: int number of cols for terminal

    Returns:
        N/A  # noqa: DAR202

    Raises:
        N/A

    """
    import fcntl
    import termios

    TIOCSWINSZ = getattr(termios, "TIOCSWINSZ", -2146929561)
    # Note, assume ws_xpixel and ws_ypixel are zero.
    s = struct.pack("HHHH", rows, cols, 0, 0)
    fcntl.ioctl(fd, TIOCSWINSZ, s)


class PtyProcess:
    def __init__(self, pid: int, fd: int) -> None:
        """
        This class represents a process running in a pseudoterminal.

        The main constructor is the `spawn` method.

        Args:
            pid: integer value of pid
            fd: integer value of fd

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        _make_eof_intr()  # Ensure _EOF and _INTR are calculated
        self.pid = pid
        self.fd = fd
        readf = io.open(fd, "rb", buffering=0)
        writef = io.open(fd, "wb", buffering=0, closefd=False)
        self.fileobj = io.BufferedRWPair(readf, writef)  # type: ignore

        self.terminated = False
        self.closed = False
        self.exitstatus: Optional[int] = None
        self.signalstatus: Optional[int] = None
        # status returned by os.waitpid
        self.status: Optional[int] = None
        self.flag_eof = False
        # Used by close() to give kernel time to update process status.
        # Time in seconds.
        self.delayafterclose = 0.1
        # Used by terminate() to give kernel time to update process status.
        # Time in seconds.
        self.delayafterterminate = 0.1

    @classmethod
    def spawn(cls: Type[PtyProcessType], spawn_command: List[str]) -> PtyProcessType:
        """Start the given command in a child process in a pseudo terminal.

        This does all the fork/exec type of stuff for a pty, and returns an
        instance of PtyProcess.
        """
        # Note that it is difficult for this method to fail.
        # You cannot detect if the child process cannot start.
        # So the only way you can tell if the child process started
        # or not is to try to read from the file descriptor. If you get
        # EOF immediately then it means that the child is already dead.
        # That may not necessarily be bad because you may have spawned a child
        # that performs some task; creates no stdout output; and then dies.

        import fcntl
        import pty
        import resource
        from pty import CHILD, STDIN_FILENO

        spawn_executable = which(spawn_command[0])
        if spawn_executable is None:
            raise SSHNotFound("ssh executable not found!")
        spawn_command[0] = spawn_executable

        # [issue #119] To prevent the case where exec fails and the user is
        # stuck interacting with a python child process instead of whatever
        # was expected, we implement the solution from
        # http://stackoverflow.com/a/3703179 to pass the exception to the
        # parent process
        # [issue #119] 1. Before forking, open a pipe in the parent process.
        exec_err_pipe_read, exec_err_pipe_write = os.pipe()

        pid, fd = pty.fork()

        # Some platforms must call setwinsize() and setecho() from the
        # child process, and others from the master process. We do both,
        # allowing IOError for either.
        if pid == CHILD:
            try:
                _setwinsize(fd=STDIN_FILENO, rows=24, cols=80)
            except IOError as err:
                if err.args[0] not in (errno.EINVAL, errno.ENOTTY):
                    raise

            # [issue #119] 3. The child closes the reading end and sets the
            # close-on-exec flag for the writing end.
            os.close(exec_err_pipe_read)
            fcntl.fcntl(exec_err_pipe_write, fcntl.F_SETFD, fcntl.FD_CLOEXEC)

            # Do not allow child to inherit open file descriptors from parent,
            # with the exception of the exec_err_pipe_write of the pipe.
            # Impose ceiling on max_fd: AIX bugfix for users with unlimited
            # nofiles where resource.RLIMIT_NOFILE is 2^63-1 and os.closerange()
            # occasionally raises out of range error
            max_fd = min(1048576, resource.getrlimit(resource.RLIMIT_NOFILE)[0])
            pass_fds = sorted({exec_err_pipe_write})
            for pair in zip([2] + pass_fds, pass_fds + [max_fd]):
                os.closerange(pair[0] + 1, pair[1])

            try:
                os.execv(spawn_executable, spawn_command)
            except OSError as err:
                # [issue #119] 5. If exec fails, the child writes the error
                # code back to the parent using the pipe, then exits.
                tosend = f"OSError:{err.errno}:{str(err)}".encode()
                os.write(exec_err_pipe_write, tosend)
                os.close(exec_err_pipe_write)
                os._exit(os.EX_OSERR)

        # Parent
        inst = cls(pid, fd)

        # [issue #119] 2. After forking, the parent closes the writing end
        # of the pipe and reads from the reading end.
        os.close(exec_err_pipe_write)
        exec_err_data = os.read(exec_err_pipe_read, 4096)
        os.close(exec_err_pipe_read)

        # [issue #119] 6. The parent reads eof (a zero-length read) if the
        # child successfully performed exec, since close-on-exec made
        # successful exec close the writing end of the pipe. Or, if exec
        # failed, the parent reads the error code and can proceed
        # accordingly. Either way, the parent blocks until the child calls
        # exec.
        if len(exec_err_data) != 0:
            try:
                errclass, errno_s, errmsg = exec_err_data.split(b":", 2)
                exctype = getattr(builtins, errclass.decode("ascii"), Exception)

                exception = exctype(errmsg.decode("utf-8", "replace"))
                if exctype is OSError:
                    exception.errno = int(errno_s)
            except Exception:
                raise Exception("Subprocess failed, got bad error data: %r" % exec_err_data)
            else:
                raise exception

        try:
            inst.setwinsize(rows=24, cols=80)
        except IOError as err:
            if err.args[0] not in (errno.EINVAL, errno.ENOTTY, errno.ENXIO):
                raise

        return inst

    def __repr__(self) -> str:
        """
        Magic repr method for PtyProcess

        Args:
            N/A

        Returns:
            str: str repr of object

        Raises:
            N/A

        """
        return f"{type(self).__name__}(pid={self.pid}, fd={self.fd})"

    def __del__(self) -> None:
        """
        Magic delete method for PtyProcess

        This makes sure that no system resources are left open. Python only
        garbage collects Python objects. OS file descriptors are not Python
        objects, so they must be handled explicitly. If the child file
        descriptor was opened outside of this class (passed to the constructor)
        then this does not close it.

        Args:
            N/A

        Returns:
            N/A

        Raises:
            N/A

        """
        if not self.closed:
            # It is possible for __del__ methods to execute during the
            # teardown of the Python VM itself. Thus self.close() may
            # trigger an exception because os.close may be None.
            try:
                self.close()
            # which exception, shouldn't we catch explicitly .. ?
            except Exception:
                pass

    def close(self, force: bool = True) -> None:
        """This closes the connection with the child application. Note that
        calling close() more than once is valid. This emulates standard Python
        behavior with files. Set force to True if you want to make sure that
        the child is terminated (SIGKILL is sent if the child ignores SIGHUP
        and SIGINT)."""
        if not self.closed:
            self.flush()
            self.fileobj.close()  # Closes the file descriptor
            # Give kernel time to update process status.
            time.sleep(self.delayafterclose)
            if self.isalive():
                if not self.terminate(force):
                    raise PtyProcessError("Could not terminate the child.")
            self.fd = -1
            self.closed = True
            # self.pid = None

    def flush(self) -> None:
        """
        This does nothing.

        It is here to support the interface for a File-like object.

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    def read(self, size: int = 1024) -> bytes:
        """Read and return at most ``size`` bytes from the pty.

        Can block if there is nothing to read. Raises :exc:`EOFError` if the
        terminal was closed.

        Unlike Pexpect's ``read_nonblocking`` method, this doesn't try to deal
        with the vagaries of EOF on platforms that do strange things, like IRIX
        or older Solaris systems. It handles the errno=EIO pattern used on
        Linux, and the empty-string return used on BSD platforms and (seemingly)
        on recent Solaris.
        """
        try:
            s = self.fileobj.read1(size)
        except (OSError, IOError) as err:
            if err.args[0] == errno.EIO:
                # Linux-style EOF
                self.flag_eof = True
                raise EOFError("End Of File (EOF). Exception style platform.")
            raise
        if s == b"":
            # BSD-style EOF (also appears to work on recent Solaris (OpenIndiana))
            self.flag_eof = True
            raise EOFError("End Of File (EOF). Empty string style platform.")

        return s

    def write(self, bytes_to_write: bytes, flush: bool = True) -> int:
        """Write bytes to the pseudoterminal.

        Returns the number of bytes written.
        """
        n = self.fileobj.write(bytes_to_write)
        if flush:
            self.fileobj.flush()
        return n

    def eof(self) -> bool:
        """This returns True if the EOF exception was ever raised."""

        return self.flag_eof

    def terminate(self, force: bool = False) -> bool:
        """
        This forces a child process to terminate.

        It starts nicely with SIGHUP and SIGINT. If "force" is True then moves onto SIGKILL. This
        returns True if the child was terminated. This returns False if the child could not be
        terminated.

        Args:
            force: bool; force termination

        Returns:
            bool: terminate succeeded or failed

        Raises:
            N/A

        """
        if not self.isalive():
            return True
        try:
            self.kill(signal.SIGHUP)
            time.sleep(self.delayafterterminate)
            if not self.isalive():
                return True
            self.kill(signal.SIGCONT)
            time.sleep(self.delayafterterminate)
            if not self.isalive():
                return True
            self.kill(signal.SIGINT)
            time.sleep(self.delayafterterminate)
            if not self.isalive():
                return True
            if force:
                self.kill(signal.SIGKILL)
                time.sleep(self.delayafterterminate)
                if not self.isalive():
                    return True
        except OSError:
            # I think there are kernel timing issues that sometimes cause
            # this to happen. I think isalive() reports True, but the
            # process is dead to the kernel.
            # Make one last attempt to see if the kernel is up to date.
            time.sleep(self.delayafterterminate)
            if not self.isalive():
                return True
        return False

    def isalive(self) -> bool:
        """This tests if the child process is running or not. This is
        non-blocking. If the child was terminated then this will read the
        exitstatus or signalstatus of the child. This returns True if the child
        process appears to be running or False if not. It can take literally
        SECONDS for Solaris to return the right status."""

        if self.terminated:
            return False

        if self.flag_eof:
            # This is for Linux, which requires the blocking form
            # of waitpid to get the status of a defunct process.
            # This is super-lame. The flag_eof would have been set
            # in read_nonblocking(), so this should be safe.
            waitpid_options = 0
        else:
            waitpid_options = os.WNOHANG

        try:
            pid, status = os.waitpid(self.pid, waitpid_options)
        except OSError as e:
            # No child processes
            if e.errno == errno.ECHILD:
                raise PtyProcessError(
                    "isalive() encountered condition "
                    + 'where "terminated" is 0, but there was no child '
                    + "process. Did someone else call waitpid() "
                    + "on our process?"
                )
            raise

        # I have to do this twice for Solaris.
        # I can't even believe that I figured this out...
        # If waitpid() returns 0 it means that no child process
        # wishes to report, and the value of status is undefined.
        if pid == 0:
            try:
                ### os.WNOHANG) # Solaris!
                pid, status = os.waitpid(self.pid, waitpid_options)
            except OSError as e:  # pragma: no cover
                # This should never happen...
                if e.errno == errno.ECHILD:
                    raise PtyProcessError(
                        "isalive() encountered condition "
                        + "that should never happen. There was no child "
                        + "process. Did someone else call waitpid() "
                        + "on our process?"
                    )
                raise

            # If pid is still 0 after two calls to waitpid() then the process
            # really is alive. This seems to work on all platforms, except for
            # Irix which seems to require a blocking call on waitpid or select,
            # so I let read_nonblocking take care of this situation
            # (unfortunately, this requires waiting through the timeout).
            if pid == 0:
                return True

        if pid == 0:
            return True

        if os.WIFEXITED(status):
            self.status = status
            self.exitstatus = os.WEXITSTATUS(status)
            self.signalstatus = None
            self.terminated = True
        elif os.WIFSIGNALED(status):
            self.status = status
            self.exitstatus = None
            self.signalstatus = os.WTERMSIG(status)
            self.terminated = True
        elif os.WIFSTOPPED(status):
            raise PtyProcessError(
                "isalive() encountered condition "
                + "where child process is stopped. This is not "
                + "supported. Is some other process attempting "
                + "job control with our child pid?"
            )
        return False

    def kill(self, sig: int) -> None:
        """Send the given signal to the child application.

        In keeping with UNIX tradition it has a misleading name. It does not
        necessarily kill the child unless you send the right signal. See the
        :mod:`signal` module for constants representing signal numbers.
        """

        # Same as os.kill, but the pid is given for you.
        if self.isalive():
            os.kill(self.pid, sig)

    def setwinsize(self, rows: int = 24, cols: int = 80) -> None:
        """
        Set window size.

        This will cause a SIGWINCH signal to be sent to the child. This does not change the physical
        window size. It changes the size reported to TTY-aware applications like vi or curses --
        applications that respond to the SIGWINCH signal.

        Args:
            rows: int number of rows for terminal
            cols: int number of cols for terminal

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        return _setwinsize(self.fd, rows, cols)



from pathlib import Path

import pytest

import scrapli
from scrapli.driver.base_network_driver import PrivilegeAction, PrivilegeLevel
from scrapli.driver.core import IOSXEDriver
from scrapli.exceptions import UnknownPrivLevel
from scrapli.response import MultiResponse, Response

TEST_DATA_DIR = f"{Path(scrapli.__file__).parents[1]}/tests/test_data/"


def test_check_kwargs_comms_prompt_pattern():
    with pytest.warns(UserWarning) as warn:
        conn = IOSXEDriver(host="localhost", comms_prompt_pattern="something")
    assert (
        conn.comms_prompt_pattern
        == r"(^[a-z0-9.\-_@()/:]{1,63}>$)|(^[a-z0-9.\-_@/:]{1,63}#$)|(^[a-z0-9.\-_@/:]{1,63}\(conf[a-z0-9.\-@/:\+]{"
        "0,32}\\)#$)"
    )
    assert (
        str(warn[0].message) == "\n***** `comms_prompt_pattern` found in kwargs! "
        "*****************************************\n`comms_prompt_pattern` is ignored (dropped) when using network "
        "drivers. If you wish to modify the patterns for any network driver sub-classes, please do so by modifying "
        "or providing your own `privilege_levels`.\n***** `comms_prompt_pattern` found in kwargs! "
        "*****************************************"
    )


def test_generate_comms_prompt_pattern(sync_cisco_iosxe_conn):
    assert (
        sync_cisco_iosxe_conn.comms_prompt_pattern
        == "(^[a-z0-9.\\-_@()/:]{1,63}>$)|(^[a-z0-9.\\-_@/:]{1,63}#$)|(^[a-z0-9.\\-_@/:]{1,63}\\(conf[a-z0-9.\\-@/:\\+]{0,32}\\)#$)"
    )


def test_build_priv_map(sync_cisco_iosxe_conn):
    assert sync_cisco_iosxe_conn._priv_map == {
        "exec": ["exec"],
        "privilege_exec": ["exec", "privilege_exec"],
        "configuration": ["exec", "privilege_exec", "configuration"],
    }


def test_update_privilege_levels(sync_cisco_iosxe_conn):
    sync_cisco_iosxe_conn.privilege_levels["scrapli"] = PrivilegeLevel(
        pattern=r"^weirdpatterndude$",
        name="scrapli",
        previous_priv="",
        deescalate="",
        escalate="",
        escalate_auth=False,
        escalate_prompt="",
    )
    sync_cisco_iosxe_conn.update_privilege_levels()
    assert (
        sync_cisco_iosxe_conn.comms_prompt_pattern
        == r"(^[a-z0-9.\-_@()/:]{1,63}>$)|(^[a-z0-9.\-_@/:]{1,63}#$)|(^[a-z0-9.\-_@/:]{1,63}\(conf[a-z0-9.\-@/:\+]{0,32}\)#$)|(^weirdpatterndude$)"
    )
    assert sync_cisco_iosxe_conn._priv_map == {
        "exec": ["exec"],
        "privilege_exec": ["exec", "privilege_exec"],
        "configuration": ["exec", "privilege_exec", "configuration"],
        "scrapli": ["scrapli"],
    }


def test_determine_current_priv_exceptions(sync_cisco_iosxe_conn):
    with pytest.raises(UnknownPrivLevel):
        sync_cisco_iosxe_conn._determine_current_priv("!!!!thisissoooowrongggg!!!!!!?!")


def test_determine_current_priv(sync_cisco_iosxe_conn):
    current_priv = sync_cisco_iosxe_conn._determine_current_priv("execprompt>")
    assert len(current_priv) == 1
    assert current_priv[0] == "exec"


def test_get_privilege_level_name_exceptions(sync_cisco_iosxe_conn):
    with pytest.raises(UnknownPrivLevel) as exc:
        sync_cisco_iosxe_conn._get_privilege_level_name(requested_priv="boo")
    assert (
        str(exc.value)
        == "Requested privilege level `boo` not a valid privilege level of `IOSXEDriver`"
    )


def test_get_privilege_level_name(sync_cisco_iosxe_conn):
    resolved_privilege_level = sync_cisco_iosxe_conn._get_privilege_level_name(
        requested_priv="exec"
    )
    assert resolved_privilege_level == "exec"


def test_update_response(sync_cisco_iosxe_conn):
    response = Response("localhost", "some input")
    sync_cisco_iosxe_conn.textfsm_platform = "racecar"
    sync_cisco_iosxe_conn.genie_platform = "tacocat"
    sync_cisco_iosxe_conn._update_response(response)
    assert response.textfsm_platform == "racecar"
    assert response.genie_platform == "tacocat"


def test_register_configuration_session(sync_cisco_iosxe_conn):
    with pytest.raises(NotImplementedError) as exc:
        sync_cisco_iosxe_conn.register_configuration_session(session_name="boo")
    assert str(exc.value) == "Configuration sessions not supported for `IOSXEDriver`"


def test_pre_escalate():
    with pytest.warns(UserWarning) as warn:
        conn = IOSXEDriver(host="localhost", comms_prompt_pattern="something", auth_secondary="")
        conn._pre_escalate(escalate_priv=conn.privilege_levels["privilege_exec"])
    assert (
        str(warn[1].message)
        == "\n***** Privilege escalation generally requires an `auth_secondary` password, but none is set! \nscrapli will try to escalate privilege without entering a password but may fail.\nSet an `auth_secondary` password if your device requires a password to increase privilege, otherwise ignore this message.\n***** Privilege escalation generally requires an `auth_secondary` password, but none is set! "
    )


def test_pre_acquire_priv(sync_cisco_iosxe_conn):
    resolved_priv, map_to_desired_priv = sync_cisco_iosxe_conn._pre_acquire_priv(
        desired_priv="configuration"
    )
    assert resolved_priv == "configuration"
    assert map_to_desired_priv == ["exec", "privilege_exec", "configuration"]


def test_process_acquire_priv_escalate(sync_cisco_iosxe_conn):
    privilege_action, priv = sync_cisco_iosxe_conn._process_acquire_priv(
        resolved_priv="configuration",
        map_to_desired_priv=["exec", "privilege_exec", "configuration"],
        current_prompt="csr1000v#",
    )
    assert privilege_action == PrivilegeAction.ESCALATE


def test_process_acquire_priv_deescalate(sync_cisco_iosxe_conn):
    privilege_action, priv = sync_cisco_iosxe_conn._process_acquire_priv(
        resolved_priv="exec",
        map_to_desired_priv=["exec", "privilege_exec", "configuration"],
        current_prompt="csr1000v#",
    )
    assert privilege_action == PrivilegeAction.DEESCALATE


def test_process_acquire_priv_no_action(sync_cisco_iosxe_conn):
    privilege_action, priv = sync_cisco_iosxe_conn._process_acquire_priv(
        resolved_priv="privilege_exec",
        map_to_desired_priv=["exec", "privilege_exec", "configuration"],
        current_prompt="csr1000v#",
    )
    assert privilege_action == PrivilegeAction.NO_ACTION


def test_pre_send_config_exceptions(sync_cisco_iosxe_conn):
    with pytest.raises(TypeError) as exc:
        sync_cisco_iosxe_conn._pre_send_config(config=["boo"])
    assert (
        str(exc.value)
        == "`send_config` expects a single string, got <class 'list'>, to send a list of configs use the `send_configs` method instead."
    )


def test_pre_send_config(sync_cisco_iosxe_conn):
    split_configs = sync_cisco_iosxe_conn._pre_send_config(
        config="interface loopback0\ndescription scrapli is neat"
    )
    assert isinstance(split_configs, list)
    assert split_configs[0] == "interface loopback0"
    assert split_configs[1] == "description scrapli is neat"


def test_post_send_config(sync_cisco_iosxe_conn):
    response_one = Response("localhost", "some input", failed_when_contains=["something"])
    response_one._record_response(result=b"greatsucccess")
    response_two = Response("localhost", "some input")
    response_two._record_response(result=b"alsosucess")
    multi_response = MultiResponse()
    multi_response.append(response_one)
    multi_response.append(response_two)
    unified_response = sync_cisco_iosxe_conn._post_send_config(
        config="interface loopback0\ndescription scrapli is neat", multi_response=multi_response
    )
    assert unified_response.failed is False
    assert unified_response.result == "greatsucccess\nalsosucess"


def test_pre_send_configs_exceptions(sync_cisco_iosxe_conn):
    with pytest.raises(TypeError) as exc:
        sync_cisco_iosxe_conn._pre_send_configs(configs="boo")
    assert (
        str(exc.value)
        == "`send_configs` expects a list of strings, got <class 'str'>, to send a single configuration line/string use the `send_config` method instead."
    )


def test_pre_send_configs_standard(sync_cisco_iosxe_conn):
    resolved_privilege_level, failed_when_contains = sync_cisco_iosxe_conn._pre_send_configs(
        configs=[]
    )
    assert resolved_privilege_level == "configuration"
    assert failed_when_contains == [
        "% Ambiguous command",
        "% Incomplete command",
        "% Invalid input detected",
        "% Unknown command",
    ]


def test_pre_send_configs_user_defined(sync_cisco_iosxe_conn):
    resolved_privilege_level, failed_when_contains = sync_cisco_iosxe_conn._pre_send_configs(
        configs=[], privilege_level="privilege_exec", failed_when_contains=["somethingneat"]
    )
    assert resolved_privilege_level == "privilege_exec"
    assert failed_when_contains == ["somethingneat"]


def test_post_send_configs(sync_cisco_iosxe_conn):
    response_one = Response("localhost", "some input", failed_when_contains=["something"])
    response_one._record_response(result=b"greatsucccess")
    multi_response = MultiResponse()
    multi_response.append(response_one)
    updated_responses = sync_cisco_iosxe_conn._post_send_configs(responses=multi_response)
    assert updated_responses[0].textfsm_platform == "cisco_iosxe"
    assert updated_responses[0].genie_platform == "iosxe"


def test_pre_send_configs_from_file_exceptions(sync_cisco_iosxe_conn):
    with pytest.raises(TypeError) as exc:
        sync_cisco_iosxe_conn._pre_send_configs_from_file(file=["boo"])
    assert (
        str(exc.value)
        == "`send_configs_from_file` expects a string path to a file, got <class 'list'>"
    )


def test_pre_send_configs_from_file(sync_cisco_iosxe_conn):
    configs = sync_cisco_iosxe_conn._pre_send_configs_from_file(
        file=f"{TEST_DATA_DIR}/files/vrnetlab_key"
    )
    assert configs[0] == "-----BEGIN OPENSSH PRIVATE KEY-----"
    assert configs[-1] == "-----END OPENSSH PRIVATE KEY-----"
